#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Pyearocom module for reading and processing of EBAS NASA Ames files

For details on the file format see `here <https://ebas-submit.nilu.no/
Submit-Data/Getting-started>`__
"""
import os
import numpy as np
import pandas as pd
from collections import OrderedDict as od
from datetime import datetime
from pyaerocom._lowlevel_helpers import str_underline, dict_to_str
from pyaerocom.exceptions import TimeZoneError
from pyaerocom import logger

class NasaAmesReadError(IOError):
    pass

class NasaAmesVariableError(AttributeError):
    pass

class EbasColDef(dict):
    """Dict-like object for EBAS NASA Ames column definitions
    
    Note
    ----
    The meta attribute name 'unit' can also be accessed using the CF attr name
    'units'
    
    Attributes
    ----------
    name : str
        column name 
    unit : str
        unit of data in column (if applicable)
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    flag_col : int
        column number of flag column that corresponds to this data column (only
        relevant if :attr:`is_var` is True)
        
    Parameters
    ----------
    name : str
        column name 
    is_var : bool
        True if column corresponds to variable data, False if not
    is_flag : bool
        True, if column corresponds to Flag column, False if not
    unit : :obj:`str`, optional
        unit of data in column (if applicable)
    flag_col : :obj:`str`, optional
        ``name`` of flag column that corresponds to this data colum (only
        relevant if :attr:`is_var` is True)
    """
    def __init__(self, name, is_var, is_flag, unit="1"):
        self.name = name
        self.unit = unit
        self.is_var = is_var
        self.is_flag = is_flag
        self.flag_col = None
        
    def get_wavelength_nm(self):
        """Try to access wavelength information in nm (as float)"""
        if not 'wavelength' in self:
            raise KeyError('Column variable {} does not contain wavelength '
                           'information'.format(self.name))
        elif not 'nm' in self.wavelength:
            raise NotImplementedError('Wavelength definition is not in nm')
        return float(self.wavelength.split('nm')[0].strip())
        
    def to_dict(self, ignore_keys=['is_var', 'is_flag', 'flag_col', 
                                   'wavelength_nm']):
        d = {}
        for k, v in self.items():
            if k in ignore_keys:
                continue
            elif k == 'unit':
                k ='units'
            d[k] = v
        return d
          
    def __getitem__(self, key): # add support for units attr (CF standard name)
        if key == 'units':
            key = 'unit'
        return super(EbasColDef, self).__getitem__(key)
    
    def __getattr__(self, key):
        return self[key]
    
    def __setattr__(self, key, val):
        self[key] = val
    
    def to_str(self):
        s = '{}_{}'.format(self.name, self.unit)
        if 'wavelength' in self:
            s += '_{}nm'.format(self.get_wavelength_nm())
        if 'matrix' in self:
            s += '_{}'.format(self.matrix)
        if 'statistics' in self:
            s += '_{}'.format(self.statistics)
        if 'data_level' in self:
            s += '_L{}'.format(self.data_level)
        return s
            
            
    def __repr__(self):
        s = "{}: ".format(type(self).__name__)
        for k, v in self.items():
            s += "{}={}, ".format(k, v)
        return s
    
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "{}\n{}\n".format(head, len(head)*"-")
        for k, v in self.items():
            s += "{}: {}\n".format(k, v)
        return s
    
def _readline_ref_and_revision(line):
    spl = line.strip().split()
    basedate = np.datetime64(datetime.strptime('{}{}{}'.format(spl[0],
                                               spl[1], spl[2]), "%Y%m%d"), 's')
    rev = np.datetime64(datetime.strptime('{}{}{}'.format(spl[3],
                                               spl[4], spl[5]), "%Y%m%d"), 's')
    return [basedate, rev]

class NasaAmesHeader(object):
    """Header class for Ebas NASA Ames file
    
    Note
    ----
    Is used in :class:`EbasNasaAmesFile` and should not be used directly.
    
    Attributes
    -----------
    
    """
    _NUM_FIXLINES = 13
    _HEAD_ROWS_MANDATORY = [0,5,8,9,10,11]
    
    #conversion methods for first 13 header lines of
    CONV_STR = lambda l : str(l.strip())
    CONV_PI = lambda l : '; '.join([x.strip() for x in l.split(';')])
    CONV_MULTIINT = lambda l : [int(x) for x in l.strip().split()]
    CONV_MULTIFLOAT = lambda l : [float(x) for x in l.strip().split()]
    CONV_INT = lambda l : int(l.strip())
    CONV_FLOAT = lambda l : float(l.strip())
    _STARTDATE_FMT = "%Y%m%d%H%M%S"

    # the parameters that are always in the beginning of the file
    _H_FIXLINES_YIELD = [["num_head_lines", "num_head_fmt"], #line 1
                         "data_originator", #line 2
                         "sponsor_organisation", #3
                         "submitter", #4
                         "project_association", #5
                         ["vol_num", "vol_totnum"], #6
                         ["ref_date", "revision_date"], #7
                         "freq", #8
                         "descr_time_unit", #9
                         "num_cols_dependent", #10
                         "mul_factors", #11
                         "vals_invalid", #12
                         "descr_first_col", #13
                         ]
    # conversion methods are defined above
    _H_FIXLINES_CONV = [CONV_MULTIINT, #1 -> yields 2
                        CONV_PI, #2
                        CONV_STR, #3
                        CONV_STR, #4
                        CONV_STR, #5
                        CONV_MULTIINT, #6
                        _readline_ref_and_revision, #7 lambda l : [x.strip() for x in l.strip().split("     ")], #7
                        CONV_FLOAT, #8
                        CONV_STR, #9
                        CONV_INT, #10
                        CONV_MULTIFLOAT, #11
                        CONV_MULTIFLOAT, #12
                        CONV_STR] #13
    
    def __init__(self, **kwargs):
        self._head_fix = od(num_head_lines = np.nan,
                            num_head_fmt = np.nan,
                            data_originator = "",
                            sponsor_organisation = "",
                            submitter = "",
                            project_association = "",
                            vol_num = np.nan,
                            vol_totnum = np.nan,
                            ref_date = np.nan,
                            revision_date = np.nan,
                            freq = np.nan,
                            descr_time_unit = "",
                            num_cols_dependent = np.nan,
                            mul_factors = [],
                            vals_invalid = [],
                            descr_first_col = "")
        self._var_defs = []
        self._meta = od()
        self.update(**kwargs)
        
    @property
    def head_fix(self):
        """Dictionary containing fixed header info (that is always available)"""
        return self._head_fix
    
    @property
    def var_defs(self):
        """List containing column variable definitions
        
        List index is column index in file and value is instance of 
        :class:`EbasColDef`
        """
        return self._var_defs
        
    @property
    def meta(self):
        """Meta data dictionary (specific for this file)"""
        return self._meta
        
    def update(self, **kwargs):
        for k, v in kwargs.items():
            try:
                self[k] = v
            except:
                logger.warning("Invalid attribute: {}".format(k))
                    
            
    def __getattr__(self, key):
        if key in self._head_fix:
            return self._head_fix[key]
        elif key in self._meta:
            return self._meta[key]
        else:
            raise AttributeError("Invalid attribute: {}".format(key))
     
    def __getitem__(self, key):
        return self.__getattr__(key)
    
    def __setitem__(self, key, val):
        if key in self._head_fix:
            self._head_fix[key] = val
        else:
            self._meta[key] = val
    
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "{}\n{}\n".format(head, len(head)*"-")
        s += dict_to_str(self._head_fix)
        s += "\n\n{}".format(str_underline("Column variable definitions",
                                           indent=3))
        for item in self._var_defs:
            s += "\n   {}".format(repr(item))
        s += "\n\n{}".format(str_underline("EBAS meta data",
                                             indent=3))
        s += dict_to_str(self.meta)
        
        return s

class EbasFlagCol(object):
    """Simple helper class to decode and interpret EBAS flag columns
    
    Attributes
    ----------
    info : EbasColDef
        EBAS column definition information for flag column (from file header)
    raw_data : ndarray
        raw flag column (containing X-digit floating point numbers)
    decoded : ndarray
        
    """
    from pyaerocom import const
    FLAG_INFO = const.EBAS_FLAG_INFO
    def __init__(self, raw_data, info, interpret_on_init=True):
        if not raw_data.ndim == 1:
            raise AttributeError("Need one dimensional numpy array for flag "
                                 "column")
        self.info = info
        self.raw_data = raw_data
        
        self._decoded = None
        self._valid = None
        
        if interpret_on_init:
            self.decode()
            
    @property
    def decoded(self):
        """Nx3 numpy array containing decoded flag columns"""
        if self._decoded is None:
            self.decode()
        return self._decoded
    
    @property
    def valid(self):
        """Boolean array specifying valid and invalid measurements"""
        if self._valid is None:
            self.decode()
        return self._valid
    
    def decode(self):
        """Decode raw flag column"""
        flags = np.zeros((len(self.raw_data), 3)).astype(int)
        mask = self.raw_data.astype(bool) # rmooves all points that are 0, i.e. that contain no flag (valid measurements)
        valid = np.ones_like(self.raw_data).astype(bool)
        not_ok = self.raw_data[mask] 
        if len(not_ok) > 0:
            _decoded = []
            _valid = []
            for flag in not_ok:
                item = "{:.9f}".format(flag).split(".")[1]
                vals = [int(item[:3]), int(item[3:6]), int(item[6:9])]
                _invalid = False
                for val in vals:
                    if val == 100:
                        _invalid = False
                        break # since all other flags are irrelevant if 100 is flagged
                    elif val in self.FLAG_INFO.valid and not self.FLAG_INFO.valid[val]: #is invalid
                        _invalid = True
                        
                _decoded.append(vals)
                _valid.append(not _invalid)
            
            flags[mask] = np.asarray(_decoded)
            valid[mask] = _valid

        self._valid = valid
        self._decoded = flags
        
class EbasNasaAmesFile(NasaAmesHeader):
    """EBAS NASA Ames file interface
    
    Class interface for reading and processing of EBAS NASA Ames file
    
    Attributes
    ----------
    time_stamps : ndarray
        array containing datetime64 objects with timestamps
    flags : dict
        dictionary containing :class:`EbasFlagCol` objects for each column
        containing flags
        
    Parameters
    ----------
    file : :obj:`str`, optional
        EBAS NASA Ames file. if valid file path, then the file is read on 
        init (please note following options for import)
    only_head : bool
        read only file header
    replace_invalid_nan : bool
        replace all invalid values in the table by NaNs. The invalid values for
        each dependent data column are identified based on the information in 
        the file header.
    convert_timestamps : bool
        compute array of numpy datetime64 timestamps from numeric timestamps
        in data
    evaluate_flags : bool
        if True, all flags in all flag columns are decoded from floating 
        point representation to 3 integers, e.g. 
        0.111222333 -> 111 222 333
    quality_check : bool
        perform quality check after import (for details see 
        :func:`_quality_check`)
    **kwargs 
        optional input args that are passed to init of :class:`NasaAmesHeader`
        base class
    
    """
    TIMEUNIT2SECFAC = dict(days = 3600*24,
                           Days = 3600*24)
    
    ERR_LOW_STATS = 'percentile:15.87'
    ERR_HIGH_STATS = 'percentile:84.13'
    def __init__(self, file=None, only_head=False, replace_invalid_nan=True,
                 convert_timestamps=True, evaluate_flags=False, 
                 quality_check=True, **kwargs):
        super(EbasNasaAmesFile, self).__init__(**kwargs)
        self._data_header = [] #Header line of data block
        self._data = [] #data block
        
        self.time_stamps = None
        
        self.flag_col_info = od()
        
        self.file = None
        
        if file is not None:
            if not os.path.exists(file):
                raise IOError("File {} does not exists".format(file))
            self.read_file(file, only_head, replace_invalid_nan, 
                           convert_timestamps, evaluate_flags,
                           quality_check)
 
    @property
    def data(self):
        """2D numpy array containing data table"""
        return self._data
    
    @property
    def data_header(self):
        return self._data_header
    
    @property
    def shape(self):
        """Shape of data array"""
        return self.data.shape
    
    @property
    def col_num(self):
        """Number of columns in table"""
        return self.num_cols_dependent + 1
    
    @property
    def col_names(self):
        """Column names of table"""
        #cols = [x["name"] for x in self.var_defs]
        return [x["name"] for x in self.var_defs]
    
    @property
    def col_names_vars(self):
        """Names of all columns that are flagged as variables"""
        return [x.name for x in self.var_defs if x.is_var]
    
    @property
    def col_nums_vars(self):
        """Column index number of all variables"""
        return [i for (i, item) in enumerate(self.var_defs) if item.is_var]

    @property
    def base_date(self):
        """Base date of data as numpy.datetime64[s]"""
        if not "timezone" in self.meta:
            raise AttributeError("Fatal: could not infer base date. Timezone "
                                 "is not available in file header")
        if not self.timezone.lower() == "utc":
            raise TimeZoneError("Timezones other than UTC are not yet supported")
        return self.ref_date
# =============================================================================
#         return np.datetime64(datetime.strptime(self.startdate, "%Y%m%d%H%M%S"),
#                              's')
# =============================================================================
    
    
    @property
    def time_unit(self):
        """Time unit of data"""
        return self.descr_time_unit.split()[0].strip()
    
    @staticmethod
    def numarr_to_datetime64(basedate, num_arr, mulfac_to_sec):
        """Convert array of numerical timestamps into datetime64 array
        
        Parameters
        ----------
        basedate : datetime64
            reference date
        num_arr : ndarray
            numerical time stamps relative to basedate
        mulfac_to_sec : float
            multiplicative factor to convert numerical values to unit of 
            seconds
        Returns
        -------
        ndarray
            array containing timestamps as datetime64 objects
        """
        totnum = len(num_arr)
        if totnum == 0:
            raise AttributeError("No data available in file")
# =============================================================================
#         elif totnum == 1:
#             num_arr = np.asarray([num_arr])
# =============================================================================
        return basedate + (num_arr * mulfac_to_sec).astype("timedelta64[s]")
    
    def get_colnames_unique(self):
        """Create a list of unique column names"""
        names = []
        for i, colinfo in enumerate(self.var_defs):
            s = colinfo.to_str()
            if not s in names:
                names.append(s)
            else:
                s += '(Col{})'.format(i)
                names.append(s)
        return names
       
    def _find_col_matches(self, var_name=None, wavelength_nm=None,
                          statistics=None, data_level=None,  **kwargs):
        """Find indices of columns that match input constraints
        
        Parameters
        ----------
        var_name : :obj:`str`, optional
            EBAS variable name (e.g. aerosol_light_scattering_coefficient). 
            If specified, only columns corresponding to this variable name 
            will be extracted into the dataframe
        wavelength_nm : :obj:`int`, :obj:`tuple`, optional
            wavelength (or wavelength range -> list or tuple input) in nm. If
            specified, only columns containing wavelength dependent data as
            specified are extracted and put into the Dataframe
        statistics : :obj:`str`, optional
            specify column statistics (e.g. arithmetic mean)
        data_level : float
            data level of column
        
        """
        try: #statistics is defined globally
            stats_glob = self.statistics 
        except: #statistics is not defined globally
            stats_glob = None
            
        if isinstance(wavelength_nm, str):
            raise ValueError("Wavelength needs to be integer or float or range"
                             "(i.e. 2-element tuple or list)")
        
        try: 
            if not len(wavelength_nm) == 2:
                raise Exception
            wvl_rng = True
        except:
            wvl_rng = False
        
        if statistics is not None:
            if isinstance(statistics, str): # input for statistics is string, else ...
                stats_str = True
            else: # ... input for statistics is assumed to be tuple or list of strings
                stats_str = False
        if data_level is not None:
            try:
                data_level_glob = float(self.data_level)
                if not data_level_glob:
                    raise Exception #is undefined
            except:
                data_level_glob = None
        cols = []
        for i, colinfo in enumerate(self.var_defs):
            if var_name is not None and not colinfo.name == var_name:
                continue
            if wavelength_nm is not None:
                try: #column contains wavalength information
                    wvl = colinfo.get_wavelength_nm()
                    if wvl_rng:
                        if not wavelength_nm[0] <= wvl <= wavelength_nm[1]:
                            continue
                    else:
                        if not wvl == wavelength_nm:
                            continue
                except: #column does not contain wavelength information (skip)
                    continue
            if statistics is not None:
                if 'statistics' in colinfo:
                    col_stats = colinfo['statistics']
                else:
                    col_stats = stats_glob
                if col_stats is None: # no statistics info available for column
                    continue
                if stats_str and not col_stats == statistics:
                    continue
                elif not col_stats in statistics:
                    continue
            if data_level is not None:
                if 'data_level' in colinfo:
                    col_lev = float(colinfo.data_level)
                else:
                    col_lev = data_level_glob
                if col_lev is None:
                    continue
                if not col_lev == data_level:
                    continue
            cols.append(i)
        return cols
    
    def _find_col(self, var_name, statistics='arithmetic mean', **kwargs):
        idx = self._find_col_matches(var_name, statistics=statistics, 
                                     **kwargs)
        #cols = self.get_colnames_unique()
        if len(idx) == 0:
            raise ValueError('No matches could be found for input specs')
        elif len(idx) > 1:
                
            msg = ('Multiple matches found for input variable, please specify '
                   'further constraints. The following column matches were '
                   'found:\n')
            for _idx in idx:
                msg += '{}\n'.format(self.var_defs[_idx])
        
            raise ValueError(msg)
        return idx[0]
    
    def to_dataframe(self, var_name=None, wavelength_nm=None,
                     statistics=None):
        """Convert table to dataframe
        
        Parameters
        ----------
        var_name : :obj:`str`, optional
            EBAS variable name (e.g. aerosol_light_scattering_coefficient). 
            If specified, only columns corresponding to this variable name 
            will be extracted into the dataframe
        wavelength_nm : :obj:`int`, :obj:`tuple`, optional
            wavelength (or wavelength range -> list or tuple input) in nm. If
            specified, only columns containing wavelength dependent data as
            specified are extracted and put into the Dataframe
        statistics : :obj:`str`, optional
            specify column statistics (e.g. arithmetic mean)
        
        """
        cols = self.get_colnames_unique()
        if all([x == None for x in (var_name, wavelength_nm, statistics)]):
            return pd.DataFrame(data=self.data,
                                index=self.time_stamps,
                                columns=cols)
        else:
            matches = self._find_col_matches(var_name, wavelength_nm,
                                             statistics)
            if len(matches) == 0:
                raise ValueError('No column matches could be found for input '
                                 'specs')
            _cols = []
            for idx in matches:
                _cols.append(cols[idx])
            return pd.DataFrame(data=self.data[:, matches],
                                index=self.time_stamps,
                                columns=_cols)
            
    def plot_var(self, var_name, statistics='arithmetic mean', ax=None, 
                 style=None, **kwargs):
        """Plot time series of one column
        
        If percentiles are available, they will be plotted as shaded area
        
        Parameters
        ----------
        var_name : str
            EBAS variable name
        statistics : str
            statistics specifications
        """
        idx = self._find_col(var_name, statistics='arithmetic mean', **kwargs)
        colinfo = self.var_defs[idx]
        try :
            idx_err_low = self._find_col(var_name, 
                                         statistics=self.ERR_LOW_STATS,
                                         **kwargs)
            idx_err_high = self._find_col(var_name, 
                                          statistics=self.ERR_HIGH_STATS,
                                          **kwargs)
        except:
            idx_err_high, idx_err_low = None, None
            print('Could not find percentile columns')
            
        s = pd.Series(self.data[:, idx], self.time_stamps)
        if s.isnull().all():
            raise ValueError('All values are NaN in column {} are NaN'.format(repr(colinfo)))
        
        if ax is None:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots()
        s.plot(style=style, ax=ax)
        return ax
        
    def assign_flagcols(self):
        _prev = 0
        for (idx, item) in enumerate(self.var_defs):
            if item.is_flag:
                for _idx in range(_prev, idx):
                    self.var_defs[_idx].flag_col=idx
                _prev = idx + 1
        
            
    def init_flags(self, evaluate=True):
        """Decode flag columns and store info in :attr:`flags`
        """
        for (idx, item) in enumerate(self.var_defs):
            if item.is_flag:
                data = self.data[:, idx]
                flag = EbasFlagCol(raw_data=data,
                                   info=item, 
                                   interpret_on_init=evaluate)
                self.flag_col_info[idx] = flag
    
    
    def _set_invalid_flags_nan_col(self, colnum):
        raise NotImplementedError
        info = self.var_defs[colnum]
        flag_col = info.flag_col
        if flag_col is None:
            raise ValueError('No flag column is assigned to data column {}\n{}'
                             .format(colnum, info))
        
        
    def set_invalid_flags_nan(self, colnum=None):
        """Use flag column information to identify and remove invalid measurements"""
        if len(self.flag_col_info) == 0:
            self.init_flags()
        raise NotImplementedError
        
        
                
    def compute_time_stamps(self):
        """Compute time stamps from first two data columns"""
        if not self.var_defs[0].unit == self.var_defs[1].unit:
            raise NasaAmesReadError('2nd column is not endtime or does not '
                                    'have the same unit as 1st column (starttime)')
        offs = self.base_date
        unit = self.time_unit
        if not unit in self.TIMEUNIT2SECFAC:
            raise ValueError("Invalid unit for temporal resolution: {}".format(unit))
        mulfac = self.TIMEUNIT2SECFAC[unit]
        
        start = self.numarr_to_datetime64(offs, self.data[:,0], mulfac)
        stop = self.numarr_to_datetime64(offs, self.data[:,1], mulfac)
            
        self.time_stamps = start + (stop - start)*.5
        return (start, stop)
    
    def _quality_check(self):
        msgs = ""
        if not len(self.data_header) == len(self.var_defs):
            msgs += ("Mismatch between variable definitions in header and "
                     "number of data columns in table\n")
        if not "timezone" in self.meta:
            msgs += ("Timezone not defined in metadata")
        if msgs:
            raise AttributeError("Quality check failed. Messages: {}".format(msgs))
        
    def read_header(self, nasa_ames_file, quality_check=True):
        self.read_file(nasa_ames_file, only_head=True, 
                       quality_check=quality_check)
        
    def read_file(self, nasa_ames_file, only_head=False, 
                  replace_invalid_nan=True, convert_timestamps=True, 
                  evaluate_flags=False, quality_check=False):
        """Read NASA Ames file
        
        Parameters
        ----------
        nasa_ames_file : str
            EBAS NASA Ames file
        only_head : bool
            read only file header
        replace_invalid_nan : bool
            replace all invalid values in the table by NaNs. The invalid values for
            each dependent data column are identified based on the information in 
            the file header.
        convert_timestamps : bool
            compute array of numpy datetime64 timestamps from numeric timestamps
            in data
        evaluate_flags : bool
            if True, all data columns get assigned their corresponding flag
            column, the flags in all flag columns are decoded from floating 
            point representation to 3 integers, e.g. 0.111222333 -> 111 222 333
            and if input ```replace_invalid_nan==True```, then the invalid 
            measurements in each column are replaced with NaN's.
        quality_check : bool
            perform quality check after import (for details see 
            :func:`_quality_check`)
        """
        logger.info("Reading NASA Ames file:\n{}".format(nasa_ames_file))
        lc = 0 #line counter
        dc = 0 #data block line counter
        mc = 0 #meta block counter
        END_VAR_DEF = np.nan #will be set (info stored in header)
        IN_DATA = False 
        data = []
        _insert_invalid = None
        self.file = nasa_ames_file
        for line in open(nasa_ames_file):
            #print(lc, _NUM_FIXLINES, line)
            if IN_DATA: #in data block (end of file)
                if dc == 0:
                    logger.debug(line)
                try:
                    data.append(tuple([float(x.strip()) for x in line.strip().split()]))
                    #data.append([float(x.strip()) for x in line.strip().split()])
                except Exception as e:
                    data.append(_insert_invalid)
                    logger.warning("Failed to read data row {}. "
                                   "Error msg: {}".format(dc, repr(e)))
                dc += 1
            elif lc < self._NUM_FIXLINES: #in header section (before column definitions)
                try:
                    val = self._H_FIXLINES_CONV[lc](line)
                    attr = self._H_FIXLINES_YIELD[lc]
                    if isinstance(attr, list):    
                        for i, attr_id in enumerate(attr):
                            self[attr_id] = val[i]
                    else:
                        self[attr] = val
                except Exception as e:
                    msg = ("Failed to read header row {}.\n{}\n" 
                           "Error msg: {}".format(lc, line, repr(e)))
                    if lc in self._HEAD_ROWS_MANDATORY:
                        raise NasaAmesReadError("Fatal: {}".format(msg))
                    else:
                        logger.warning(msg)
            else: # behind header section and before data definition (contains column defs and meta info)
                if mc == 0: # still in column definition
                    END_VAR_DEF = self._NUM_FIXLINES + self.num_cols_dependent - 1
                    NUM_HEAD_LINES = self.num_head_lines
                    try:
                        self.var_defs.append(self._read_vardef_line(line))
                    except Exception as e:
                        logger.warning(repr(e))
                           
                elif lc < END_VAR_DEF:
                    self.var_defs.append(self._read_vardef_line(line))
                    
                elif lc == NUM_HEAD_LINES - 1:
                    IN_DATA = True
                    self._data_header = h = [x.strip() for x in line.split()]
                    #append information of first two columns to variable 
                    #definition array.
                    self._var_defs.insert(0, EbasColDef(name=h[0],
                                                        is_flag=False,
                                                        is_var=False,
                                                        unit=self.time_unit))
                    self._var_defs.insert(1, EbasColDef(name=h[1],
                                                        is_flag=False,
                                                        is_var=False,
                                                        unit=self.time_unit))
                    if only_head:
                        return
                    logger.debug("REACHED DATA BLOCK")
                    _insert_invalid = tuple([np.nan]*self.col_num)
                    
                    
                #elif lc > self._NUM_FIXLINES + 3:
                elif lc >= END_VAR_DEF + 2:
                    try:
                        name, val = line.split(":")
                        key = name.strip().lower().replace(" ", "_")
                        self.meta[key] = val.strip()
                    except Exception as e:
                        logger.warning("Failed to read line no. {}.\n{}\n"
                              "Error msg: {}\n".format(lc, line, repr(e)))
                else:
                    logger.debug("Ignoring line no. {}: {}".format(lc, line)) 
                mc += 1
            lc += 1
        
        
        
        data = np.asarray(data) 
        
        data[:, 1:] = data[:, 1:] * np.asarray(self.mul_factors)
        
        self._data = data
        if replace_invalid_nan:
            dep_dat = data[:, 1:]
            for i, val in enumerate(np.floor(self.vals_invalid)):
                col = dep_dat[:, i]
                cond = np.floor(col) == val
                col[cond] = np.nan
                dep_dat[:, i] = col

            data[:, 1:] = dep_dat
        self._data = data
        
        if convert_timestamps:
            self.compute_time_stamps()
            
        self.assign_flagcols()
        self.init_flags(evaluate=evaluate_flags)
        
        if quality_check:
            self._quality_check()
            
    def _read_vardef_line(self, line_from_file):
        """Import variable definition line from NASA Ames file"""
        spl = [x.strip() for x in line_from_file.split(",")]
        name = spl[0]
        if not len(spl) > 1:
            unit = ''
        else:
            unit = spl[1]
        data = EbasColDef(name=name,
                          is_flag=True,
                          is_var=False,
                          unit=unit)
        
        if not "numflag" in name:    
            data.is_var = True
            data.is_flag = False
            for item in spl[2:]:
                if "=" in item:
                    # e.g. wavelength=550nm
                    sub = item.split("=")
                    if len(sub) == 2:
                        idf, val = [x.strip() for x in sub]
                        data[idf.lower().replace(' ', '_')] = val
                    else:
                        logger.warning("Could not interpret part of column "
                                       "definition in EBAS NASA Ames file: "
                                       "{}".format(item))
                else: #unit
                    logger.warning("Failed to interpret {}".format(item))
        
        return data
    
    def _data_short_str(self):
        if len(self.data) == 0:
            s = "No data available"
        else:
            s =  str(self.data)
            shape = self.data.shape
            s += "\nColnum: {}".format(shape[1])
            s += "\nTimestamps: {}".format(shape[0])
        return s            
    
    def print_col_info(self):
        """Print information about individual columns"""
        for (idx, coldef) in enumerate(self.var_defs):
            print("Column {}\n{}".format(idx, coldef))
            
    def __str__(self):
        s = super(EbasNasaAmesFile, self).__str__()
        s += "\n\n{}".format(str_underline("Data", indent=3))
        s += "\n{}".format(self._data_short_str())
        return s
    
if __name__=="__main__":
    DIR_MC = "/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/EBASMultiColumn/data/data/"
    FILES_MC = ['US0035R.20150826222034.20160117041718.nephelometer..pm10.5mn.1h.US06L_TSI_3563_BND_pm10.US06L_scat_coef.lev2.nas']
    
    
    file_mc = os.path.join(DIR_MC, FILES_MC[0])
        
    mc = EbasNasaAmesFile(file_mc)
    print(mc)
    
    #alert = EbasNasaAmesFile(DIR_MC + FILES_MC[1])
    #print(alert)
    
    #df = alert.to_dataframe()
    
    #idx = alert._find_col_matches('aerosol_absorption_coefficient',
    #                              statistics='arithmetic mean')
    
    
    
    #alert.plot_var('aerosol_absorption_coefficient', data_level=2)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 24 11:07:06 2018

@author: jonasg
"""
import os
try:
    from ConfigParser import ConfigParser
except: 
    from configparser import ConfigParser

from pyaerocom import __dir__, logger
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.io import EbasSQLRequest

class EbasVarInfo(BrowseDict):
    """EBAS I/O variable information for Aerocom
    
    See `variables.ini <https://github.com/metno/pyaerocom/blob/master/
    pyaerocom/data/variables.ini>`__ file for an overview of currently available 
    default variables.
    
    Attributes
    ----------
    var_name : str
        Aerocom variable name
    
    """
    def __init__(self, var_name="abs550aer", init=True, **kwargs):
        self.var_name = var_name
        
        #: aliases
        self.aliases = []
        
        #: old variable name
        self.old_name = None
        
        #: list of variable / component names (EBAS side)
        self.component = None
        
        #: list of matrix names (EBAS side, optional)
        self.matrix = None
        
        #: list of instrument names (EBAS side, optional)
        self.instrument = None
        
        #: list containing variable statistics info (EBAS side, optional)
        self.statistics = None
        
        #: list of additional variable required for retrieval of this variable
        self.requires = None
        
        #: scale factor for conversion to Aerocom units
        self.scale_factor = 1
        
        #imports default information and, on top, variable information (if 
        # applicable)
        if init:
            self.parse_from_ini(var_name) 
    
    @staticmethod
    def PROVIDES_VARIABLES():
        data = EbasVarInfo.open_config()
        return [k for k in data.keys()]
        
    @staticmethod
    def open_config():
        fpath = os.path.join(__dir__, "data", "ebas_config.ini")
        if not os.path.exists(fpath):
            raise IOError("Ebas config file could not be found: %s"
                          %fpath)
        conf_reader = ConfigParser()
        conf_reader.read(fpath)
        return conf_reader
    
    def _check_aliases(self, varname, conf_reader):
        for section, item in conf_reader.items():
            if 'aliases' in item:
                if varname in [x.strip() for x in item['aliases'].split(',')]:
                    logger.warning('Found alias match ({}) for variable {}, '
                                   'Note that searching for aliases slows down '
                                   'things, thus, please consider using the '
                                   'actual aerocom variable '
                                   'name'.format(varname, section))
                    return section
        raise IOError('No alias match could be found for variable {}'.format(varname))
        
    def parse_from_ini(self, var_name=None, conf_reader=None):
        """Import information about default region
        
        Parameters
        ----------
        var_name : str
            strind ID of region (must be specified in `regions.ini <https://
            github.com/metno/pyaerocom/blob/master/pyaerocom/data/regions.ini>`__ 
            file)
        conf_reader : ConfigParser
            open config parser object
            
        Returns
        -------
        bool
            True, if default could be loaded, False if not
        
        Raises
        ------
        IOError
            if regions.ini file does not exist

        
        """
        if conf_reader is None:
            conf_reader = self.open_config()
        
        if not var_name in conf_reader:
            try:
                var_name = self._check_aliases(var_name, conf_reader)
            except Exception as e:
                raise IOError("No variable information  found for {} "
                              "(including aliases): Error: {}".format(var_name,
                               repr(e)))
        var_info = conf_reader[var_name]
        for key in self.keys():
            if key in var_info:
                val = var_info[key]
                if key in ('var_name', 'old_name') :
                    self[key] = val
                elif key =='scale_factor':
                    self[key] = float(val.split('#')[0].strip())
                else:
                    self[key] = list(dict.fromkeys([x for x in val.split(',')]))
        
        if self.old_name is not None:
            self.aliases.append(self.old_name)
    
    def to_dict(self):
        """Convert into dictionary"""
        d = {}
        for k, v in self.items():
            if k == 'unit':
                k = 'units'
            if v is not None:
                d[k] = v
        return d
    
    def get_all_components(self):
        """Get list of all components"""
        return get_all_components(self.var_name)
    
    def make_sql_request(self, **constraints):
        """Create an SQL request for the specifications in this object
        
        Parameters
        ----------
        **constraints
            request constraints deviating from default. For details on 
            parameters see :class:`EbasSQLRequest`
            
        Returns
        -------
        EbasSQLRequest
            the SQL request object that can be used to retrieve corresponding 
            file names using instance of :func:`EbasFileIndex.get_file_names`.
        """
        variables = self.get_all_components()
# =============================================================================
#         if self.component is not None:
#             variables.extend(self.component)
#         if self.requires is not None:
#             for aux_var in self.requires:
#                 aux_info = EbasVarInfo(aux_var)
#                 if aux_info.component is not None:
#                     variables.extend(aux_info.component)
#                 elif aux_info.requires is not None:
#                     variables.extend(aux_info.requires)
#         
#         # remove duplicates
#         variables = list(set(variables))
# =============================================================================
        if len(variables) == 0:
            raise AttributeError('At least one component (Ebas variable name) '
                             'must be specified for retrieval of variable '
                             '{}'.format(self.var_name))
    
        # default request
        req = EbasSQLRequest(variables=variables, matrices=self.matrix,
                              instrument_types=self.instrument,
                              statistics=self.statistics)
        
            
        req.update(**constraints)
        return req
        
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        for k, v in self.items():
                s += "\n%s: %s" %(k,v)
        return s

def get_all_components(var_name, varlist=None):
    """Get all EBAS components required to read a certain variable
    
    Parameters
    ----------
    var_name : str
        AeroCom variable name
    varlist : list, optional
        list of components already inferred (this function runs recursively).
        
    Returns
    -------
    list
        list of components required to read / compute input AeroCom variable
    """
    if varlist is None:
        varlist = []
    aux_info = EbasVarInfo(var_name)
    if aux_info.component is not None:
        varlist.extend(aux_info.component)
    if aux_info.requires is not None:
        for aux_var in aux_info.requires:
            varlist = get_all_components(aux_var, varlist)
    return list(set(varlist))

def check_all_variables():
    """Helper function that checks all EBAS variables against SQL database
    
    For all variables, see file ``ebas_config.ini`` in data directory
    
    Raises
    ------
    AttributeError
        if one of the variable definitions in the ini file is not according to
        requirements
    """
    from pyaerocom.io import EbasFileIndex
    all_vars = EbasVarInfo.PROVIDES_VARIABLES()
    db = EbasFileIndex()
    db_info = {'component'     :   db.ALL_VARIABLES,
               'matrix'        :   db.ALL_MATRICES,
               'instrument'    :   db.ALL_INSTRUMENTS,
               'statistics'    :   db.ALL_STATISTICS_PARAMS}
    
    errors = []
    checked_vars = []
    for varname in all_vars:
        info = EbasVarInfo(varname)
        print('Checking variable {}'.format(varname))
        if varname in checked_vars:
            errors.append('Variable {} is defined more than once'.format(varname))
        checked_vars.append(varname)
        for attr, items_db in db_info.items():
            vals_to_check = info[attr]
            if vals_to_check is not None:
                if not isinstance(vals_to_check, list):
                    raise AttributeError('Please check attribute {} of '
                                         'variable {}'.format(attr, varname))
                for item in vals_to_check:
                    if not item in items_db:
                        s=''
                        for compname, db_vals in db_info.items():
                            if item in db_vals:
                                s += ('\nAdditional info: ID {} was found in '
                                      'databes attr {}'.format(item, compname))
                        errors.append(("No such {} ({}) in database. Please "
                                       "check variable {}.{}".format(attr, 
                                                                     item,
                                                                     varname,
                                                                     s)))
                        
    return errors
            
if __name__=="__main__":
    print(EbasVarInfo('concso2'))
    
    print(get_all_components('ang4470dryaer'))



################################################################
# read_aeronet_invv2.py
#
# read Aeronet inversion V2 data
#
# this file is part of the pyaerocom package
#
#################################################################
# Created 20180629 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2018 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA

import os, re

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.mathutils import calc_abs550aer, calc_od550aer
from pyaerocom import StationData

class ReadAeronetInvV2(ReadAeronetBase):
    """Interface for reading Aeronet inversion V2 Level 1.5 and 2.0 data

    Parameters
    ----------
    dataset_to_read
        string specifying either of the supported datasets that are defined 
        in ``SUPPORTED_DATASETS``
    """
    #: Mask for identifying datafiles 
    _FILEMASK = '*.dubovikday'
    
    #: version log of this class (for caching)
    __version__ = "0.08_" + ReadAeronetBase.__baseversion__
    
    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_INV_V2L2_DAILY_NAME
    
    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_INV_V2L2_DAILY_NAME,
                          const.AERONET_INV_V2L15_DAILY_NAME]
    
    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_INV_V2L2_DAILY_NAME   :   'daily',
                const.AERONET_INV_V2L15_DAILY_NAME  :   'daily'}
    
    #: Mapping for dataset location for different data levels that can be 
    #: read with this interface (can be used when creating the object)
    DATA_LEVELS = {2.0      :   SUPPORTED_DATASETS[0],
                   1.5      :   SUPPORTED_DATASETS[1]}
    
    #: default variables for read method
    DEFAULT_VARS = ['ssa675aer','ssa440aer', 'ssa870aer', 'ssa1020aer',
                    'abs550aer',
                    'od550aer']
    
    #: value corresponding to invalid measurement
    NAN_VAL = -9999.
    
    #: dictionary specifying the file column names (values) for each Aerocom 
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE['ssa440aer'] = 'SSA440-T'
    VAR_NAMES_FILE['ssa675aer'] = 'SSA675-T'
    VAR_NAMES_FILE['ssa870aer'] = 'SSA870-T'
    VAR_NAMES_FILE['ssa1020aer'] = 'SSA1020-T'
    VAR_NAMES_FILE['od440aer'] = 'AOTExt440-T' #derived from inversion
    VAR_NAMES_FILE['ang4487aer'] = '870-440AngstromParam.[AOTExt]-Total' 
    VAR_NAMES_FILE['abs440aer'] = 'AOTAbsp440-T' 
    VAR_NAMES_FILE['angabs4487aer'] = '870-440AngstromParam.[AOTAbsp]' 
    
    #: OPTIONAL: dictionary specifying alternative column names for variables
    #: defined in :attr:`VAR_NAMES_FILE`. Check attribute _alt_vars_cols after
    #: running read().
    ALT_VAR_NAMES_FILE = {}
    ALT_VAR_NAMES_FILE['ssa440aer'] = ['SSA439-T', 'SSA441-T', 'SSA438-T', 
                                       'SSA437-T', 'SSA442-T']
    ALT_VAR_NAMES_FILE['ssa675aer'] = ['SSA676-T', 'SSA673-T', 'SSA674-T', 
                                       'SSA669-T', 'SSA677-T', 'SSA668-T', 
                                       'SSA672-T']
    ALT_VAR_NAMES_FILE['ssa870aer'] = ['SSA871-T', 'SSA869-T', 'SSA868-T', 
                                       'SSA873-T', 'SSA867-T', 'SSA872-T']
    ALT_VAR_NAMES_FILE['ssa1020aer'] = ['SSA1022-T', 'SSA1016-T', 'SSA1018-T']
    ALT_VAR_NAMES_FILE['od440aer'] = ['AOTExt439-T', 'AOTExt441-T', 
                                      'AOTExt438-T', 'AOTExt437-T',
                                      'AOTExt442-T']
    ALT_VAR_NAMES_FILE['abs440aer'] = ['AOTAbsp439-T', 'AOTAbsp441-T',
                                       'AOTAbsp438-T', 'AOTAbsp437-T',
                                       'AOTAbsp442-T']
    
    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {'abs550aer'     :   ['abs440aer',
                                         'angabs4487aer'],
                    'od550aer'      :   ['od440aer',
                                        'ang4487aer']}
                    
    #: Functions that are used to compute additional variables (i.e. one 
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {'abs550aer'     :   calc_abs550aer,
                'od550aer'      :   calc_od550aer}
    
    #: dictionary specifying the file column names (values) for each 
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    META_NAMES_FILE['data_quality_level'] = 'DATA_TYPE'
    META_NAMES_FILE['date'] = 'Date(dd-mm-yyyy)'
    META_NAMES_FILE['time'] = 'Time(hh:mm:ss)'
    META_NAMES_FILE['day_of_year'] = 'Julian_Day'
    
    #: List of variables that are provided by this dataset (will be extended 
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE.keys())

    def __init__(self, dataset_to_read=None, level=None):
        super(ReadAeronetInvV2, self).__init__(dataset_to_read)
        if level is not None:
            self.change_data_level(level)
    
    def change_data_level(self, level):
        """Change level of Inversion data
        
        Parameters
        ----------
        level :obj:`float` or :obj:`int`, 
            data level (choose from 1.5 or 2)
        
        Raises
        ------
        ValueError
            if input level is not available
        """
        if not level in self.DATA_LEVELS:
            raise ValueError('Invalid input for level, please choose '
                             'from {}'.format(self.DATA_LEVELS.keys()))
        super(ReadAeronetInvV2, self).__init__(self.DATA_LEVELS[level])
        
    def read_file(self, filename, vars_to_retrieve=None,
                  vars_as_series=False):
        """Read Aeronet file containing results from v2 inversion algorithm

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : list
            list of str with variable names to read
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects
            
        Returns
        -------
        StationData
            dict-like object containing results
        

        Example
        -------
        >>> import pyaerocom.io as pio
        >>> obj = pio.read_aeronet_invv2.ReadAeronetInvV2()
        >>> files = obj.get_file_list()
        >>> filedata = obj.read_file(files[0])
        """
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)
       
        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []
            
        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            data_out[var] = []
        
        # Iterate over the lines of the file
        self.logger.info("Reading file {}".format(filename))
    
        with open(filename, 'rt') as in_file:
            #get rid of the first com,a seperated string element...
            c_dummy = ','.join(in_file.readline().strip().split(',')[1:])
            # re.split(r'=|\,',c_dummy)
            i_dummy = iter(re.split(r'=|\,', c_dummy.rstrip()))
            dict_loc = dict(zip(i_dummy, i_dummy))

            data_out['latitude'] = float(dict_loc['lat'])
            data_out['longitude'] = float(dict_loc['long'])
            data_out['altitude'] = float(dict_loc['elev'])
            data_out['station_name'] = dict_loc['Locations']
            data_out['PI'] = dict_loc['PI']
            data_out['PI_email'] = dict_loc['Email']
            
            data_out['ts_type'] = self.TS_TYPE

            #skip next two lines
            self.logger.info('Skipping line:\n{}'.format(in_file.readline()))
            self.logger.info('Skipping line:\n{}'.format(in_file.readline()))
            
            col_index_str = in_file.readline()
            if col_index_str != self._last_col_index_str:
                self.logger.info("Header has changed, reloading col_index map")
                self._update_col_index(col_index_str)
            col_index = self.col_index
            
            # dependent on the station, some of the required input variables
            # may not be provided in the data file. These will be ignored
            # in the following list that iterates over all data rows and will
            # be filled below, with vectors containing NaNs after the file 
            # reading loop
            vars_available = {}
            col_names = {}
            for var in vars_to_read:
                if var in col_index:
                    idx = col_index[var]
                    vars_available[var] = idx
                    col_names[var] = self._last_col_order[idx]
                else:
                    self.logger.warning("Variable {} not available in file {}"
                                        .format(var, os.path.basename(filename)))
            data_out['col_names'] = col_names
            for line in in_file:
                # process line
                dummy_arr = line.strip().split(self.COL_DELIM)

                # This uses the numpy datestring64 functions that i.e. also 
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index['date']].split(':')
                datestring = '-'.join([year, month, day])
                datestring = 'T'.join([datestring, dummy_arr[col_index['time']]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])
                data_out['dtime'].append(np.datetime64(datestring))
                
                for var in self.META_NAMES_FILE:
                    val = dummy_arr[col_index[var]]
                    try:
                        # e.g. lon, lat, altitude
                        val = float(val)
                    except:
                        pass
                    data_out[var].append(val)

                # copy the data fields that are available (rest will be filled
                # below)
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL: 
                        val = np.nan
                    data_out[var].append(val)
        
        # convert all lists to numpy arrays
        data_out['dtime'] = np.asarray(data_out['dtime'])
        
        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])
            
        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out['dtime'])) * np.nan
            data_out[var] = array
        
        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)
        
        if vars_as_series:        
            for var in (vars_to_read + vars_to_compute):
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], 
                                              index=data_out['dtime'])
                else:
                    del data_out[var]
            
        return data_out

if __name__=="__main__":
    import matplotlib.pyplot as plt
    plt.close('all')
    read = ReadAeronetInvV2()
    read.verbosity_level = 'warning'
    
    #data = read.read(read.PROVIDES_VARIABLES)
    
    data  = read.read_first_file()
    
    wavelengths = [440, 675, 870, 1020]
    vals = [np.nanmean(data.ssa440aer),
            np.nanmean(data.ssa675aer),
            np.nanmean(data.ssa870aer),
            np.nanmean(data.ssa1020aer)]
    
    plt.plot(wavelengths, vals, '--og')
    
    avgssa550aer = np.interp(550, wavelengths, vals)
    
    plt.plot(550, avgssa550aer, ' xb')
    print(data)
    
    udat = read.read(last_file=10)
    print(udat.metadata)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for _lowlevel_helpers.py module of pyaerocom
"""

import numpy.testing as npt
import pyaerocom.mathutils as utils

def test_exponent():
    """Test method :func:`exponent` of :mod:`pyaerocom.utils`"""
    nominal = [-2, 0, 2]
    vals = [utils.exponent(.01),
            utils.exponent(4),
            utils.exponent(234)]
    npt.assert_array_equal(nominal, vals)
    
if __name__ == '__main__':
    test_exponent()
    
    



from .aerocom_evaluation import AerocomEvaluation
from .trends_evaluation import TrendsEvaluation
from .helpers import ModelConfigEval, ObsConfigEval
from .web_naming_conventions import VAR_MAPPING

from .cli import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
import numpy.testing as npt
import numpy as np
import os
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom.io.read_aeronet_invv3 import ReadAeronetInvV3

@lustre_unavail
def test_load_berlin():
    dataset = ReadAeronetInvV3()
    files = dataset.find_in_file_list('*Berlin*')
    assert len(files) == 1
    assert os.path.basename(files[0]) == '19930101_20190914_Berlin_FUB.all'
    data = dataset.read_file(files[0],
                             vars_to_retrieve=['abs550aer'])
    
    test_vars = ['abs440aer', 'angabs4487aer', 'abs550aer']
    assert all([x in data for x in test_vars])
    
    # more than 100 timestamps
    assert all([len(data[x]) > 100 for x in test_vars])
    
    assert isinstance(data['dtime'][0], np.datetime64)
    t0 = data['dtime'][0]
    
    assert t0 == np.datetime64('2014-07-07T12:00:00')
    
    
    first_vals = [np.nanmean(data[var]) for var in test_vars]
    
    nominal = [0.014629, 0.908436, 0.012112]
    npt.assert_allclose(actual=first_vals, desired=nominal, rtol=TEST_RTOL)
    
    
if __name__=="__main__":
    test_load_berlin()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
General helper methods for the pyaerocom library.
"""
from pyaerocom import const
import re
from pyaerocom.time_config import (PANDAS_FREQ_TO_TS_TYPE, 
                                   TS_TYPE_TO_PANDAS_FREQ)
from pyaerocom.exceptions import TemporalResolutionError

class TsType(object):
    VALID = const.GRID_IO.TS_TYPES
    FROM_PANDAS = PANDAS_FREQ_TO_TS_TYPE
    TO_PANDAS = TS_TYPE_TO_PANDAS_FREQ
    
    TS_MAX_VALS = {'hourly' : 24,
                   'daily'  : 7,
                   'weekly' : 4,
                   'monthly': 12}
    
    def __init__(self, val):
        self._mulfac = 1.0
        self._val = None
        
        self.val = val
    
    @property
    def mulfac(self):
        """Multiplication factor of frequency"""
        return self._mulfac
    
    @property
    def base(self):
        """Base string (without multiplication factor, cf :attr:`mulfac`)"""
        return self._val
    
    @property
    def name(self):
        """Name of ts_type (Wrapper for :attr:`val`)"""
        return self.val
    
    @property
    def val(self):
        """Value of frequency (string type), e.g. 3daily"""
        if self._mulfac != 1:
            return '{}{}'.format(self._mulfac, self._val)
        return self._val
    
    @val.setter
    def val(self, val):
        ival=1
        if val[-1].isdigit():
            raise TemporalResolutionError('Invalid input for TsType: {}'
                                          .format(val))
        elif val[0].isdigit():
            ivalstr = re.findall('\d+', val)[0]
            val = val.split(ivalstr)[-1]
            ival = int(ivalstr)
        if not val in self.VALID:
            try:
                val = self._from_pandas(val)
            except TemporalResolutionError:
                raise TemporalResolutionError('Invalid input. Need any valid '
                                              'ts_type: {}'
                                              .format(self.VALID))
        if val in self.TS_MAX_VALS and ival != 1:
            if ival > self.TS_MAX_VALS[val]:
                raise TemporalResolutionError('Invalid input for ts_type {}{}. '
                                              'Interval factor {} exceeds '
                                              'maximum allowed for {}, which '
                                              'is: {}'
                                              .format(ival, val, ival, val, 
                                                      self.TS_MAX_VALS[val]))
        self._val = val
        self._mulfac = ival
        
    @staticmethod
    def infer(self, datetime_index):
        """Infer resolution based on input datateime_index
        
        Uses :func:`pandas.infer_freq` to infer frequency
        """
        raise NotImplementedError
    
    def to_pandas(self):
        """Convert ts_type to pandas frequency string"""
        freq = self.TO_PANDAS[self._val]
        if self._mulfac == 1:
            return freq
        return '{}{}'.format(self._mulfac, freq)
        
    def _from_pandas(self, val):
        if not val in self.FROM_PANDAS:
            raise TemporalResolutionError('Invalid input: {}, need pandas '
                                          'frequency string'
                                          .format(val))
        return self.FROM_PANDAS[val]
    
    def __lt__(self, other):
        if self.val == other.val:
            return False
        
        idx_this = self.VALID.index(self._val)
        idx_other = self.VALID.index(other._val)
        if not idx_this == idx_other: #they have a different freq string
            return idx_this > idx_other
        #they have the same frequency string but different _mulfac attributes
        return self._mulfac > other._mulfac
    
    def __le__(self, other):
        return True if (self.__eq__(other) or self.__lt__(other)) else False
    
    def __eq__(self, other):
        return other.val == self.val
    
    def __call__(self):
        return self.val
    
    def __str__(self):
        return self.val
    
    def __repr__(self):
        return self.val
    
if __name__=="__main__":

    
    daily = TsType('daily')
    monthly = TsType('monthly')
    
    print('Monthly < daily:', monthly < daily)
    print('Monthly == daily:', monthly == daily)
    print('Daily == daily:', daily==daily)
    print('Monthly <= daily:', monthly <= daily)
    print('daily >= monthly:', daily   >= monthly)
    print('daily > monthly:', daily > monthly)
    print('Monthly >= daily:', monthly >= daily)
    
    hourly = TsType('hourly')
    hourly5 = TsType('5hourly')

    print(hourly)
    print(hourly5)
    
    print('hourly == 5hourly:', hourly==hourly5)
    print('hourly > 5hourly:', hourly>hourly5)
# =============================================================================
#     
#     class Num(object):
#         def __init__(self, val):
#             self.val = val
#             
#         def __lt__(self, other):
#             print('Other is smaller than this')
#             return self.val < other.val
#         
#     print(Num(3) < Num(5))
# =============================================================================



#!/usr/bin/env python3

################################################################
# readgridded.py
#
# model data reading class
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171030 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

#Copyright (C) 2017 met.no
#Contact information:
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA

from glob import glob
import os
from collections import OrderedDict as od
import numpy as np
import pandas as pd
import iris

from pyaerocom import const, print_log, logger
from pyaerocom.variable import Variable, is_3d
from pyaerocom.io.aux_read_cubes import (compute_angstrom_coeff_cubes,
                                         multiply_cubes,
                                         subtract_cubes)
from pyaerocom.helpers import to_pandas_timestamp, get_highest_resolution
from pyaerocom.exceptions import (DataCoverageError,
                                  DataQueryError,
                                  DataSourceError,
                                  FileConventionError,
                                  IllegalArgumentError, 
                                  TemporalResolutionError,
                                  VarNotAvailableError)

from pyaerocom.io.fileconventions import FileConventionRead
from pyaerocom.io import AerocomBrowser
from pyaerocom.io.iris_io import load_cubes_custom, concatenate_iris_cubes
from pyaerocom.io.helpers import add_file_to_log
from pyaerocom.griddeddata import GriddedData


class ReadGridded(object):
    """Class for reading gridded files based on network or model ID
    
    Note
    ----
    The reading only works if files are stored using a valid file naming 
    convention. See package data file `file_conventions.ini <http://
    aerocom.met.no/pyaerocom/config_files.html#file-conventions>`__ for valid
    keys. You may define your own fileconvention in this file, if you wish.
    
    Attributes
    ----------
    data_id : str
        string ID for model or obsdata network (see e.g. Aerocom interface map
        plots lower left corner)
    data : GriddedData
        imported data object 
    data_dir : str
        directory containing result files for this model
    start : pandas.Timestamp
        start time for data import
    stop : pandas.Timestamp
        stop time for data import
    file_convention : FileConventionRead
        class specifying details of the file naming convention for the model
    files : list
        list containing all filenames that were found.
        Filled, e.g. in :func:`ReadGridded.get_model_files`
    from_files : list
        List of all netCDF files that were used to concatenate the current 
        data cube (i.e. that can be based on certain matching settings such as
        var_name or time interval). 
    ts_types : list
        list of all sampling frequencies (e.g. hourly, daily, monthly) that 
        were inferred from filenames (based on Aerocom file naming convention) 
        of all files that were found
    vars : list
        list containing all variable names (e.g. od550aer) that were inferred 
        from filenames based on Aerocom model file naming convention
    years : list
        list of available years as inferred from the filenames in the data 
        directory.
        
    Parameters
    ----------
    data_id : str
        string ID of model (e.g. "AATSR_SU_v4.3","CAM5.3-Oslo_CTRL2016")
    start : :obj:`pandas.Timestamp` or :obj:`str`, optional
        desired start time of dataset (note, that strings are passed to 
        :class:`pandas.Timestamp` without further checking)
    stop : :obj:`pandas.Timestamp` or :obj:`str`, optional
        desired stop time of dataset (note, that strings are passed to 
        :class:`pandas.Timestamp` without further checking)
    file_convention : str
        string ID specifying the file convention of this model (cf. 
        installation file `file_conventions.ini <https://github.com/metno/
        pyaerocom/blob/master/pyaerocom/data/file_conventions.ini>`__)
    init : bool
        if True, the model directory is searched (:func:`search_data_dir`) on
        instantiation and if it is found, all valid files for this model are 
        searched using :func:`search_all_files`.

        
    """
    AUX_REQUIRES = {'ang4487aer'    : ['od440aer', 'od870aer'],
                    'od550gt1aer'   : ['od550aer', 'od550lt1aer'],
                    'conc*'        : ['mmr*', 'rho']}
    
    AUX_ALT_VARS = {'od440aer'  :   ['od443aer'],
                    'od870aer'  :   ['od865aer']}
    
    AUX_FUNS = {'ang4487aer'   :    compute_angstrom_coeff_cubes,
                'od550gt1aer'  :    subtract_cubes,
                'conc*'        :    multiply_cubes}
    
    _data_dir = ""
    
    VERT_ALT = {'Surface' : 'ModelLevel'}

    def __init__(self, data_id="", file_convention="aerocom3", init=True):
    
        if not isinstance(data_id, str):
            if isinstance(data_id, list):
                msg = ("Input for data_id is list. You might want to use "
                       "class ReadGriddedMulti for import?")
            else:
                msg = ("Invalid input for data_id. Need str, got: %s"
                       %type(data_id))
            raise TypeError(msg)
        
        #: data_id of gridded dataset        
        self.data_id = data_id
        
        self.logger = logger
        
        #: Dictionary containing loaded results for different variables
        self.data = od()
        
        # file naming convention. Default is aerocom3 file convention, change 
        # using self.file_convention.import_default("aerocom2"). Is 
        # automatically updated in class ReadGridded
        self.file_convention = FileConventionRead(file_convention)
        
        self.file_info = None
                
        #: List of unique Aerocom variable names that were identified from 
        #: the filenames in the data directory
        self._vars_2d = []
        self._vars_3d = []
        
        #: Cube lists for each loaded variable
        self.loaded_cubes = od()
        
        #: This object can be used to 
        self.browser = AerocomBrowser()
        
        self._aux_avail = None
        
        # these can be filled using method add_aux_compute and they will not 
        # affect global settings of the reader class
        self._aux_requires = {}
        self._aux_funs = {}
        
        self.ignore_vert_code = False
        if init and data_id:
            self.search_data_dir()
            self.search_all_files()
     
    
    @property
    def years_avail(self):
        """Available years"""
        if self.file_info is None:
            self.search_all_files()
        return sorted(self.file_info.year.unique())
    
    @property
    def years(self):
        """Available years"""
        print_log.warning(DeprecationWarning('Attr. "years" is deprecated '
                                             '(but still works). Please use '
                                             '"years_avail" instead.'))
        return self.years_avail
    
    @property
    def experiments(self):
        """List of all experiments that are available in this dataset"""
        if self.file_info is None:
            self.search_all_files()
        return sorted(self.file_info.experiment.unique())
    
    @property
    def files(self):
        """List  of data files"""
        if self.file_info is None:
            self.search_all_files()
        return [os.path.join(self.data_dir, x) for x in 
                sorted(self.file_info.filename.values)]
      
    @property
    def ts_types(self):
        """Available frequencies"""
        return self.file_info.ts_type.unique()
    
    @property
    def vars(self):
        return sorted(self._vars_2d + self._vars_3d)
    
    @property
    def vars_provided(self):
        """Variables provided by this interface"""
        v = []
        v.extend(self.vars)
        
        self._aux_avail = []
        for aux_var in self.AUX_REQUIRES.keys():
            try:
                self._get_aux_vars_and_fun(aux_var)
                if not aux_var in v:
                    v.append(aux_var)
                self._aux_avail.append(aux_var)
            except: #this auxiliary variable cannot be computed
                pass
        for aux_var in self._aux_requires.keys():
            try:
                self._get_aux_vars_and_fun(aux_var)
                if not aux_var in v:
                    v.append(aux_var)
                self._aux_avail.append(aux_var)
            except: #this auxiliary variable cannot be computed
                pass
        #v.extend(self.AUX_REQUIRES.keys())
        #also add standard names of 3D variables if not already in list
        for var in self._vars_3d:
            var = var.lower().replace('3d','')
            if not var in v:
                v.append(var)
        return v
    
    @property
    def data_dir(self):
        """Model directory"""
        dirloc = self._data_dir
        if not os.path.isdir(dirloc):
            raise IOError("Model directory for ID %s not available or does "
                          "not exist" %self.data_id)
        return dirloc
    
    @data_dir.setter
    def data_dir(self, value):
        if isinstance(value, str) and os.path.isdir(value):
            self._data_dir = value
        else:
            raise ValueError("Could not set directory: %s" %value)
    
    @property
    def file_type(self):
        """File type of data files"""
        return const.GRID_IO.FILE_TYPE
    
    @property
    def TS_TYPES(self):
        """List with valid filename encryptions specifying temporal resolution
        """
        return const.GRID_IO.TS_TYPES
    
    @property
    def start(self):
        """First available year in the dataset (inferred from filenames)
        
        Note
        ----
        This is not variable or ts_type specific, so it is not necessarily 
        given that data from this year is available for all variables in 
        :attr:`vars` or all frequencies liste in :attr:`ts_types`
        """
        if len(self.years_avail) == 0:
            raise AttributeError('No information about available years accessible'
                                 'please run method search_all_files first')
        yr = sorted(self.years_avail)[0]
        if yr == 9999:
            yr = 2222
        return to_pandas_timestamp(yr)
            
    @property
    def stop(self):
        """Last available year in the dataset (inferred from filenames)
        
        Note
        ----
        This is not variable or ts_type specific, so it is not necessarily 
        given that data from this year is available for all variables in 
        :attr:`vars` or all frequencies liste in :attr:`ts_types`
        """
        if len(self.years_avail) == 0:
            raise AttributeError('No information about available years accessible'
                                 'please run method search_all_files first')
        years = sorted(self.years_avail)
        year = years[-1]
        
        if year == 9999:
            self.logger.warning('Data contains climatology. Will be ignored '
                                'as stop time, using last year')
            if len(years) == 1:
                year = 2222
            else:
                year = years[-2]
            
        return to_pandas_timestamp('{}-12-31 23:59:59'.format(year))
    
    def has_var(self, var_name):
        """Check if variable is available
        
        Parameters
        ----------
        var_name : str
            variable to be checked
        
        Returns
        -------
        bool
        """
        if var_name in self.vars_provided or self.check_compute_var(var_name):
            return True
        return False
    
    def _get_years_to_load(self, start=None, stop=None):
        """Array containing year numbers that are supposed to be loaded
        
        Returns
        -------
        ndarray
            all years to be loaded
        """
        start_provided = False
        if start is None:
            start = self.start
            if start.year == 2222:
                return np.array([9999])
        else:
            start_provided = True
            if start == 9999:
                return np.array([9999])
            start = to_pandas_timestamp(start)
        
        if stop is None:
            if start_provided:
                stop = to_pandas_timestamp(start)
            else:
                stop = self.stop
        else:
            stop = to_pandas_timestamp(stop)
        if const.MIN_YEAR > start.year:
            print_log.warning('First available year {} of data {} is smaller '
                              'than supported first year {}.'
                              .format(start, self.data_id, 
                                             const.MIN_YEAR))
            start = const.MIN_YEAR
        if const.MAX_YEAR < stop.year:
            raise ValueError('Last available year {} of data {} is larger '
                             'than supported last year {}.'
                             .format(start, self.data_id, const.MAX_YEAR))
            stop = const.MAX_YEAR
            
        if start and stop:
            return np.arange(start.year, stop.year + 1, 1)
        
        if not self.years_avail:
            raise AttributeError("No information available for available "
                                 "years. Please run method "
                                 "search_all_files first")   
        return np.array(self.years_avail)
    
    def search_data_dir(self):
        """Search data directory based on model ID
        
        Wrapper for method :func:`search_data_dir_aerocom`
        
        Returns
        -------
        str
            data directory
        
        Raises
        ------
        IOError 
            if directory cannot be found
        """
        _dir = self.browser.find_data_dir(self.data_id)
        self.data_dir = _dir
        return _dir
            
    @staticmethod
    def _eval_data_id(data_id):
        spl = data_id.split('_')
        name, meteo = '',''
        experiment = spl[-1]
        if len(spl) > 1:
            sspl = spl[0].split('-')
            if len(sspl) > 1:
                name = sspl[0]
                meteo = sspl[-1]
        return (name, meteo, experiment)
    
    def search_all_files(self, update_file_convention=True):
        """Search all valid model files for this model
        
        This method browses the data directory and finds all valid files, that
        is, file that are named according to one of the aerocom file naming
        conventions. The file list is stored in :attr:`files`.
        
        Note
        ----
        It is presumed, that naming conventions of files in
        the data directory are not mixed but all correspond to either of the 
        conventions defined in 
        
        Parameters
        ----------
        update_file_convention : bool
            if True, the first file in `data_dir` is used to identify the
            file naming convention (cf. :class:`FileConventionRead`)
            
        Raises
        ------
        IOError
            if none of the files in the data directory follows either of the 
            available naming conventions (cf. data file *fileconventions.ini*)
        """
        result = []
        files_ignored = []
        # get all netcdf files in folder
        nc_files = glob(self.data_dir + '/*{}'.format(self.file_type))
        if update_file_convention:
            # Check if the found file has a naming according the aerocom conventions
            # and set the convention for all files (maybe this need to be 
            # updated in case there can be more than one file naming convention
            # within one model directory)
            ok = False
            for file in nc_files:
                try:
                    self.file_convention.from_file(os.path.basename(file))
                    ok = True
                    break
                except:
                    pass
            if not ok:
                raise IOError("Failed to identify file naming convention "
                              "from files in model directory for model "
                              "%s\ndata_dir: %s"
                              %(self.data_id, self.data_dir))
        
        _vars_temp = []
        _vars_temp_3d = []
        
        for _file in nc_files:
            # TODO: resolve this in a more general way...
            if 'ModelLevelAtStations' in _file:
                const.logger.info('Ignoring file {}'.format(_file))
                files_ignored.append(os.path.basename(_file))
                continue
            try:
                info = self.file_convention.get_info_from_file(_file)
                var_name = info['var_name']
                _is_3d = False
                if is_3d(var_name):
                    _vars_temp_3d.append(var_name)
                    _is_3d = True
                else:
                    _vars_temp.append(var_name)
                
                if not info["ts_type"] in self.TS_TYPES:
                    raise TemporalResolutionError('Invalid frequency {}'
                                                  .format(info["ts_type"]))
                
                name, meteo, experiment = self._eval_data_id(info['data_id']) 
                result.append([var_name, info['year'], info['ts_type'], 
                               info['vert_code'], self.data_id, 
                               name, meteo, experiment, 
                               info['is_at_stations'],
                               _is_3d, os.path.basename(_file)])
                    
            except (FileConventionError, DataSourceError, 
                    TemporalResolutionError) as e:
                msg = ("Failed to import file {}\nModel: {}\n"
                       "Error: {}".format(os.path.basename(_file), 
                                         self.data_id, repr(e)))
                logger.warning(msg)
                if const.WRITE_FILEIO_ERR_LOG:
                    add_file_to_log(_file, msg)
        
        if len(_vars_temp + _vars_temp_3d) == 0:
            raise AttributeError("Failed to extract information from filenames")
        # make sorted list of unique vars

        self._vars_2d = sorted(od.fromkeys(_vars_temp))
        self._vars_3d = sorted(od.fromkeys(_vars_temp_3d))
        
        
        header = ['var_name', 'year', 'ts_type', 'vert_code', 'data_id', 'name',
                  'meteo', 'experiment', 'is_at_stations', '3D', 'filename']
        df = pd.DataFrame(result, columns=header)
        df.sort_values(['var_name', 'year', 'ts_type', 'data_id', 'name',
                        'meteo', 'experiment', 'is_at_stations', '3D'], 
                        inplace=True)
        
        uv = df.vert_code.unique()
        if len(uv) == 1 and uv[0] == '':
            self.ignore_vert_code = True
        self.file_info = df
        
        if len(df) == 0:
            raise DataCoverageError('No files could be found for data {} and '
                                    'years range {}-{}'.format(self.data_id))
    
    def filter_files(self, var_name=None, ts_type=None, start=None, stop=None, 
                     experiment=None, vert_which=None, is_at_stations=False,
                     df=None):
        """Filter file database 
        
        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        ts_type : str
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised 
            of the available resolutions is used
        start : Timestamp or str, optional
            start time of data import 
        stop : Timestamp or str, optional
            stop time of data import
        experiment : str
            name of experiment (only relevant if this dataset contains more 
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case 
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that 
            match the query.
        
        """
        if df is None:
            df = self.file_info
            
        yrs = self._get_years_to_load(start, stop)
        year_cond = df.year.isin(yrs)
        
        if var_name is None:
            var_cond = df.var_name.isin(df.var_name.values)
        else:
            var_cond = df.var_name == var_name
        if vert_which is None:
            vert_cond = df.vert_code.isin(df.vert_code.values)
        else:
            vert_cond = df.vert_code == vert_which
        if ts_type is None:
            freq_cond = df.ts_type.isin(df.ts_type.values)
        else:
            freq_cond = df.ts_type == ts_type
        if experiment is None:
            exp_cond = df.experiment.isin(df.experiment.values)
        else:
            exp_cond = df.experiment == experiment
        
        return df.loc[(var_cond) & 
                      (year_cond) &
                      (freq_cond) &
                      (exp_cond) &
                      (vert_cond) &
                      (df.is_at_stations==is_at_stations)]
        
    def _infer_ts_type(self, df, ts_type, flex_ts_type,
                       prefer_longer):
        ts_types = df.ts_type.unique()
        
        if len(ts_types) == 1:
            # only one frequency available
            if flex_ts_type or ts_type is None or ts_types[0] == ts_type:
                # all good
                return ts_types[0]
            raise DataCoverageError('No files could be found for ts_type {}'
                                    .format(ts_type))
        highest_avail = get_highest_resolution(*ts_types)
        # there is more than one frequency available -> decision making 
        # gets more complicated
        if not flex_ts_type:
            if ts_type is None:
                return highest_avail
            elif ts_type in ts_types:
                return ts_type
            raise DataCoverageError('Failed to infer ts_type')
            
        # ts_type is flexible
        if ts_type is None:
            # initiate with highest available
            ts_type = highest_avail
        
        if not prefer_longer:
            return ts_type
        
        # ts_type is flexible and user prefers the longer period over 
        # higher resolution
        ts_type = ts_types[0]
        subset = self.filter_files(ts_type=ts_type, df=df)
        for _ts_type in ts_types[1:]:
            _subset = self.filter_files(ts_type=_ts_type, df=df)
            if len(_subset) > len(subset):
                subset = _subset
                ts_type = _ts_type
        return ts_type
    
    def filter_query(self, var_name, ts_type=None, start=None, stop=None, 
                         experiment=None, vert_which=None, 
                         is_at_stations=False, flex_ts_type=True, 
                         prefer_longer=False):
        """Filter files for read query based on input specs
        
        Parameters
        ----------
        
        
        Returns
        -------
        DataFrame 
            dataframe containing filtered dataset
        """
        if not var_name in self.file_info.var_name.values:
            raise DataCoverageError('Variable {} is not available in dataset '
                                    '{}'.format(var_name, self.data_id))
            
        subset = self.filter_files(var_name=var_name, 
                                   ts_type=None, # disregard ts_type in 1. iteration
                                   start=start, stop=stop, 
                                   experiment=experiment, 
                                   vert_which=vert_which, 
                                   is_at_stations=is_at_stations)
        if len(subset) == 0:
            if vert_which in self.VERT_ALT:
                vc =self.VERT_ALT[vert_which]
                const.print_log.warning('No files could be found for var {} and '
                                        'vert_which {} in {}. Trying to find '
                                        'alternative options'
                                        .format(var_name, vert_which, 
                                                self.data_id))
                return self.filter_query(var_name, ts_type, start, stop, 
                                             experiment, vert_which=vc, 
                                             is_at_stations=is_at_stations,
                                             flex_ts_type=flex_ts_type, 
                                             prefer_longer=prefer_longer)
            raise DataCoverageError('No files could be found')
        ts_type = self._infer_ts_type(subset, ts_type, flex_ts_type, 
                                      prefer_longer)
        subset = self.filter_files(ts_type=ts_type, start=start, 
                                   stop=stop, df=subset)
        if len(subset) == len(subset.year.unique()):
            return subset
        
        # File request could not be resolved such that every year only occurs
        # once
        msg =''
        exps = subset.experiment.unique()
        verts = subset.vert_code.unique()
        
        if len(exps) > 1:
            msg += 'Found multiple experiments. Choose from: {}'.format(exps)
        if len(verts) > 1:
            if msg:
                msg += '; '
            msg += 'Found multiple vertical codes. Choose from: {}'.format(verts)
        raise DataQueryError('Failed to uniquely identify data files for input '
                             'query. Reason: {}'.format(msg))
        
    def get_files(self, var_name, ts_type=None, start=None, stop=None, 
                  experiment=None, vert_which=None, 
                  is_at_stations=False, flex_ts_type=True, 
                  prefer_longer=False):
        """Get data files based on input specs"""
        subset = self.filter_query(var_name, ts_type, start, stop, 
                                       experiment, vert_which, 
                                       is_at_stations, flex_ts_type,
                                       prefer_longer)
        
        return self._generate_file_paths(subset)

    def _generate_file_paths(self, df=None):
        if df is None:
            df = self.file_info
        return sorted([os.path.join(self.data_dir, x) for x in df.filename.values])
    
    def get_var_info_from_files(self):
        """Creates dicitonary that contains variable specific meta information
        
        Returns
        -------
        OrderedDict
            dictionary where keys are available variables and values (for each
            variable) contain information about available ts_types, years, etc.
        """
        result = od()
        for file in self.files:
            finfo = self.file_convention.get_info_from_file(file)
            var_name = finfo['var_name']
            if not var_name in result:
                result[var_name] = var_info = od()
                for key in finfo.keys():
                    if not key == 'var_name':
                        var_info[key] = []
            else:
                var_info = result[var_name]
            for key, val in finfo.items():
                if key == 'var_name':
                    continue
                if val is not None and not val in var_info[key]:
                    var_info[key].append(val)
        # now check auxiliary variables
        for var_to_compute in self.AUX_REQUIRES.keys():
            if var_to_compute in result:
                continue
            try:
                vars_to_read = self._get_aux_vars_and_fun(var_to_compute)[0]
            except VarNotAvailableError:
                pass
            else:
                # init result info dict for aux variable
                result[var_to_compute] = var_info = od()
                first = result[vars_to_read[0]]
                # init with results from first required variable
                var_info.update(**first)
                if len(vars_to_read) > 1:
                    for info_other in vars_to_read[1:]:
                        other = result[info_other]
                        for key, info in var_info.items():
                            # compute match with other variable
                            var_info[key] = list(np.intersect1d(info,
                                                               other[key]))
                var_info['aux_vars'] = vars_to_read
                            
        return result
        
    def update(self, **kwargs):
        """Update one or more valid parameters
        
        Parameters
        ----------
        **kwargs
            keyword args that will be used to update (overwrite) valid class 
            attributes such as `data, data_dir, files`
        """
        for k, v in kwargs.items():
            if k in self.__dict__:
                self.logger.info("Updating %s in ModelImportResult for model %s"
                            "New value: %s" %(k, self.data_id, v))
                self.__dict__[k] = v
            else:
                self.logger.info("Ignoring key %s in ModelImportResult.update()" %k)
                 
    def concatenate_cubes(self, cubes):
        """Concatenate list of cubes into one cube
        
        Parameters
        ----------
        CubeList
            list of individual cubes
        
        Returns
        -------
        Cube
            Single cube that contains concatenated cubes from input list
            
        Raises 
        ------
        iris.exceptions.ConcatenateError
            if concatenation of all cubes failed 
        """        
        return concatenate_iris_cubes(cubes, error_on_mismatch=True)
    
# =============================================================================
#     def _get_aux_fun(self, var_to_compute):
#         """Get method used to compute input variable
#         
#         Parameters
#         ----------
#         var_to_compute : str
#             variable to be computed
#         
#         Returns
#         -------
#         callable
#             method that is used to compute variable
#         
#         Raises
#         ------
#         AttributeError
#             if no function is defined for that variable name
#         """
#         if var_to_compute in self.AUX_FUNS:
#             return self.AUX_FUNS[var_to_compute]
#         elif var_to_compute in self._aux_funs:
#             return self._aux_funs[var_to_compute]
#         raise AttributeError('No method found for computation of {}'
#                              .format(var_to_compute))
#         
#     def _get_aux_vars(self, var_to_compute):
#         """Helper that searches auxiliary variables for computation of input var
#         
#         Parameters
#         ----------
#         var_to_compute : str
#             one of the auxiliary variables that is supported by this interface
#             (cf. :attr:`AUX_REQUIRES`)
#         
#         Returns 
#         -------
#         list
#             list of variables that are used as input for computation method 
#             of input variable (cf. :attr:`AUX_FUNS`)
#             
#         Raises 
#         ------
#         VarNotAvailableError
#             if one of the required variables for computation is not available 
#             in the data
#         """
#         if var_to_compute in self._aux_requires:
#             vars_req = self._aux_requires[var_to_compute]
#         elif var_to_compute in self.AUX_REQUIRES:
#             vars_req = self.AUX_REQUIRES[var_to_compute]
#         
#         vars_to_read = []
#         for var in vars_req:
#             found = 0
#             if var in self.vars:
#                 found = 1
#                 vars_to_read.append(var)
#             elif var in self.AUX_ALT_VARS:
#                 for alt_var in list(self.AUX_ALT_VARS[var]):
#                     if alt_var in self.vars:
#                         found = 1
#                         vars_to_read.append(alt_var)
#                         break
#             if not found:
#                 raise VarNotAvailableError('Cannot compute {}, since {} '
#                                            '(req. for computation) is not '
#                                            'available in data'
#                                            .format(var_to_compute, 
#                                                    var))
#         return vars_to_read
# =============================================================================
    
    def _get_aux_vars_and_fun(self, var_to_compute):
        """Helper that searches auxiliary variables for computation of input var
        
        Parameters
        ----------
        var_to_compute : str
            one of the auxiliary variables that is supported by this interface
            (cf. :attr:`AUX_REQUIRES`)
        
        Returns 
        -------
        list
            list of variables that are used as input for computation method 
            of input variable (cf. :attr:`AUX_FUNS`)
            
        Raises 
        ------
        VarNotAvailableError
            if one of the required variables for computation is not available 
            in the data
        """
        if (var_to_compute in self._aux_requires and 
            var_to_compute in self._aux_funs):
            vars_req = self._aux_requires[var_to_compute]
            fun = self._aux_funs[var_to_compute]
        elif (var_to_compute in self.AUX_REQUIRES and 
              var_to_compute in self.AUX_FUNS):
            vars_req = self.AUX_REQUIRES[var_to_compute]
            fun = self.AUX_FUNS[var_to_compute]
        
        vars_to_read = []
        for var in vars_req:
            found = 0
            if var in self.vars:
                found = 1
                vars_to_read.append(var)
            elif var in self.AUX_ALT_VARS:
                for alt_var in list(self.AUX_ALT_VARS[var]):
                    if alt_var in self.vars:
                        found = 1
                        vars_to_read.append(alt_var)
                        break
            if not found:
                raise VarNotAvailableError('Cannot compute {}, since {} '
                                           '(req. for computation) is not '
                                           'available in data'
                                           .format(var_to_compute, 
                                                   var))
        return (vars_to_read, fun)
    
    def compute_var(self, var_name, start=None, stop=None, ts_type=None, 
                    experiment=None, vert_which=None, flex_ts_type=True, 
                    prefer_longer=False, vars_to_read=None, aux_fun=None, 
                    **kwargs):
        """Compute auxiliary variable
        
        Like :func:`read_var` but for auxiliary variables 
        (cf. :attr:`AUX_REQUIRES`)
        
        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        start : :obj:`Timestamp` or :obj:`str`, optional
            start time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        stop : :obj:`Timestamp` or :obj:`str`, optional
            stop time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        ts_type : str
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised 
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more 
            than one experiment)
        vert_which : str
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel)
        flex_ts_type : bool
            if True and if applicable,start=None, stop=None,
                 ts_type=None, flex_ts_type=True then another ts_type is used in case 
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that 
            match the query.
        **kwargs
            additional keyword args passed to :func:`_load_var`
            
        Returns
        -------
        GriddedData
            loaded data object
        """
        if vars_to_read is not None:
            self.add_aux_compute(var_name, vars_to_read, aux_fun)
        vars_to_read, aux_fun = self._get_aux_vars_and_fun(var_name)
        
        data = []
        # all variables that are required need to be in the same temporal
        # resolution
        try:
            ts_type = self.find_common_ts_type(vars_to_read, start, stop, 
                                               ts_type, experiment, 
                                               vert_which=vert_which,
                                               flex_ts_type=flex_ts_type)  
        except DataCoverageError:
            if not vert_which in self.VERT_ALT:
                raise
            vert_which = self.VERT_ALT[vert_which]
            
            ts_type = self.find_common_ts_type(vars_to_read, start, stop, 
                                               ts_type, experiment, 
                                               vert_which=vert_which,
                                               flex_ts_type=flex_ts_type)  
        for var in vars_to_read:
            aux_data = self._load_var(var_name=var, 
                                      ts_type=ts_type, 
                                      start=start, stop=stop,
                                      experiment=experiment,
                                      vert_which=vert_which,
                                      flex_ts_type=flex_ts_type,
                                      prefer_longer=prefer_longer,
                                      **kwargs)
            data.append(aux_data)

        cube = aux_fun(*data)
        cube.var_name = var_name
        
        data = GriddedData(cube, data_id=self.data_id, 
                           ts_type=data[0].ts_type,
                           computed=True)
        return data
    
    def find_common_ts_type(self, vars_to_read, start=None, stop=None,
                            ts_type=None, experiment=None, vert_which=None, 
                            flex_ts_type=True):
        """Find common ts_type for list of variables to be read
        
        Parameters
        ----------
        vars_to_read : list
            list of variables that is supposed to be read
        start : :obj:`Timestamp` or :obj:`str`, optional
            start time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        stop : :obj:`Timestamp` or :obj:`str`, optional
            stop time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        ts_type : str
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised 
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more 
            than one experiment)
        vert_which : str
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel)
        flex_ts_type : bool
            if True and if applicable,start=None, stop=None,
                 ts_type=None, flex_ts_type=True then another ts_type is used in case 
            the input ts_type is not available for this variable
        
        Returns
        -------
        str 
            common ts_type for input variable
            
        Raises
        ------
        DataCoverageError
            if no match can be found
            
        """
        if isinstance(vars_to_read, str):
            vars_to_read = [vars_to_read]
        
        common = self.filter_files(var_name=vars_to_read[0], 
                                   start=start,
                                   stop=stop,
                                   experiment=experiment,
                                   vert_which=vert_which).ts_type.unique()
        if len(common) == 0:
            raise DataCoverageError('Could not find any file matches for query '
                                    'and variable {}'.format(vars_to_read[0]))
        for var in vars_to_read[1:]:
            _tt = self.filter_files(var_name=var, 
                                    start=start,
                                    stop=stop,
                                    experiment=experiment,
                                    vert_which=vert_which)
            common = np.intersect1d(common, _tt.ts_type.unique())
            
        if len(common) == 0:
            raise DataCoverageError('Could not find common ts_type for '
                                    'variables {}'.format(vars_to_read))
        elif len(common) == 1:
            if ts_type is None or flex_ts_type:
                return common[0]
            elif ts_type == common[0]:
                return ts_type
            raise DataCoverageError('Could not find files with ts_type={} for '
                                    'all input variables: {}'
                                    .format(ts_type, vars_to_read))
        if ts_type is not None:
            if ts_type in common:
                return ts_type
            
        if not flex_ts_type:
            raise DataCoverageError('Could not find files with ts_type={} for '
                                    'all input variables: {}'
                                    .format(ts_type, vars_to_read))
        
        common_sorted = [x for x in const.GRID_IO.TS_TYPES if x in common]
        return common_sorted[0]
    
    def add_aux_compute(self, var_name, vars_required, fun):
        """Register new variable to be computed
        
        Parameters
        ----------
        var_name : str
            variable name to be computed
        vars_required : list
            list of variables to read, that are required to compute `var_name`
        fun : callable
            function that takes a list of `GriddedData` objects as input and 
            that are read using variable names specified by `vars_required`.
        """
        if isinstance(vars_required, str):
            vars_required = [vars_required]
        if not isinstance(vars_required, list):
            raise ValueError('Invalid input for vars_required. Need str or list. '
                             'Got: {}'.format(vars_required))
        elif not callable(fun):
            raise ValueError('Invalid input for fun. Input is not a callable '
                             'object')
        self._aux_requires[var_name] = vars_required
        self._aux_funs[var_name] = fun
    
    def check_compute_var(self, var_name):
        """Check if variable name belongs to family that can be computed
        
        For instance, if input var_name is `concdust` this method will check
        :attr:`AUX_REQUIRES` to see if there is a variable family pattern
        (`conc*`) defined that specifies how to compute these variables. If 
        a match is found, the required variables and computation method is 
        added via :func:`add_aux_compute`.
        
        Note
        ----
        BETA version, use with care
        
        Parameters
        -----------
        var_name : str
            variable name to be checked
        
        Returns
        -------
        bool
            True if match is found, else False
        """
        
        if '*' in var_name:
            from pyaerocom.exceptions import VariableDefinitionError
            raise VariableDefinitionError('Invalid variable name {}. Must not '
                                          'contain *'.format(var_name))
        if var_name in self.AUX_REQUIRES:
            return True
        elif var_name in self._aux_requires:
            return True
        import fnmatch
        patterns = [x for x in self.AUX_REQUIRES if '*' in x]
        vars_found = []
        for pattern in patterns:
            if fnmatch.fnmatch(var_name, pattern):
                vars_required = self.AUX_REQUIRES[pattern]
                for addvar in vars_required:
                    
                    if not '*' in addvar:
                        vars_found.append(addvar)
                    else:
                        _addvar = var_name
                        spl1 = pattern.split('*')
                        spl2 = addvar.split('*')
                        if len(spl1) != len(spl2):
                            raise AttributeError('variable patterns in '
                                                 'AUX_REQUIRES and corresponding '
                                                 'values (with * in name) need '
                                                 'to have the same number of '
                                                 'wildcard delimiters')
                        for i, substr in enumerate(spl1):
                            if bool(substr):
                                _addvar = _addvar.replace(substr, spl2[i])
                        vars_found.append(_addvar)
                if (len(vars_found) == len(vars_required) and
                    all([x in self.vars_provided for x in vars_found])):
                    
                    self.add_aux_compute(var_name, 
                                         vars_required=vars_found,
                                         fun=self.AUX_FUNS[pattern])
                        
                    return True
        return False
    
                
    # TODO: add from_vars input arg for computation and corresponding method
    def read_var(self, var_name, start=None, stop=None,
                 ts_type=None, experiment=None, vert_which=None, 
                 flex_ts_type=True, prefer_longer=False, 
                 aux_vars=None, aux_fun=None,
                 **kwargs):
        """Read model data for a specific variable
        
        This method searches all valid files for a given variable and for a 
        provided temporal resolution (e.g. *daily, monthly*), optionally
        within a certain time window, that may be specified on class 
        instantiation or using the corresponding input parameters provided in 
        this method.
        
        The individual NetCDF files for a given temporal period are loaded as
        instances of the :class:`iris.Cube` object and appended to an instance
        of the :class:`iris.cube.CubeList` object. The latter is then used to 
        concatenate the individual cubes in time into a single instance of the
        :class:`pyaerocom.GriddedData` class. In order to ensure that this
        works, several things need to be ensured, which are listed in the 
        following and which may be controlled within the global settings for 
        NetCDF import using the attribute :attr:`GRID_IO` (instance of
        :class:`OnLoad`) in the default instance of the 
        :class:`pyaerocom.config.Config` object accessible via 
        ``pyaerocom.const``.
        
        
        Parameters
        ----------
        var_name : str
            variable that are supposed to be read
        start : Timestamp or str, optional
            start time of data import 
        stop : Timestamp or str, optional
            stop time of data import
        ts_type : str
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised 
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more 
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case 
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that 
            match the query.
        aux_vars : list
            only relevant if `var_name` is not available for reading but needs
            to be computed: list of variables that are required to compute 
            `var_name`
        aux_fun : callable
            only relevant if `var_name` is not available for reading but needs
            to be computed: custom method for computation (cf. 
            :func:`add_aux_compute` for details)
        **kwargs
            additional keyword args parsed to :func:`_load_var`
            
        Returns
        -------
        GriddedData
            loaded data object
            
        Raises
        ------
        AttributeError
            if none of the ts_types identified from file names is valid
        VarNotAvailableError
            if specified ts_type is not supported
        """
        if aux_vars is not None:
            self.add_aux_compute(var_name, aux_vars, aux_fun)
        
        if isinstance(ts_type, dict):
            try:
                ts_type = ts_type[var_name]
            except:
                const.print_log.info('Setting ts_type to None, since input '
                                     'dict {} does not contain specification '
                                     'variable to read {}'.format(ts_type, 
                                                                  var_name))
                ts_type = None
                
        #ts_type = self._check_ts_type(ts_type)
        var_to_read = None
        do_compute = (var_name in self._aux_requires and 
                      self.check_compute_var(var_name))
        if not do_compute:
            if var_name in self.vars:
                var_to_read = var_name  
            else:
                # e.g. user asks for od550aer but files contain only 3d var od5503daer
                #if not var_to_read in self.vars: 
                for var in self._vars_3d:
                    if Variable(var).var_name == var_name:
                        var_to_read = var
                if var_to_read is None:
                    for alias in const.VARS[var_name].aliases:
                        if alias in self.vars:
                            const.print_log.info('Did not find {} field, loading '
                                                 '{} instead'.format(var_name,
                                                  alias))
                            var_to_read = alias
        
        if isinstance(vert_which, dict):
            try:
                vert_which = vert_which[var_name]
            except:
                const.print_log.info('Setting vert_which to None, since input '
                                     'dict {} does not contain specification '
                                     'variable to read {}'.format(vert_which, 
                                                                  var_name))
                vert_which = None
                
        if var_to_read is not None: # variable can be read directly
            data = self._load_var(var_name=var_to_read, 
                                  ts_type=ts_type, 
                                  start=start, stop=stop,
                                  experiment=experiment, 
                                  vert_which=vert_which,
                                  flex_ts_type=flex_ts_type, 
                                  prefer_longer=prefer_longer,
                                  **kwargs)
        
        elif self.check_compute_var(var_name):
            data = self.compute_var(var_name=var_name, 
                                    start=start, stop=stop, 
                                    ts_type=ts_type, 
                                    experiment=experiment, 
                                    vert_which=vert_which,
                                    flex_ts_type=flex_ts_type, 
                                    prefer_longer=prefer_longer)

# =============================================================================
#                                     vars_to_read=self._aux_requires[var_name],
#                                     aux_fun=self._aux_funs[var_name])
#             
# =============================================================================
        else:
            raise VarNotAvailableError("Error: variable {} not available in "
                                       "files and can also not be computed."
                                       .format(var_name))
        
        if var_name in self.data:
            self.logger.warning("Warning: Data for variable {} already exists "
                           "and will be overwritten".format(var_name))
        
        data.reader = self
        self.data[var_name] = data
        
        return data
        
                
    def read(self, vars_to_retrieve=None, start=None, stop=None, ts_type=None, 
             experiment=None, vert_which=None, flex_ts_type=True, 
             prefer_longer=False, require_all_vars_avail=False, **kwargs):
        """Read all variables that could be found 
        
        Reads all variables that are available (i.e. in :attr:`vars`)
        
        Parameters
        ----------
        vars_to_retrieve : list or str, optional
            variables that are supposed to be read. If None, all variables
            that are available are read.
        start : Timestamp or str, optional
            start time of data import 
        stop : Timestamp or str, optional
            stop time of data import 
        ts_type : str, optional
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly"). If None, prioritised 
            of the available resolutions is used
        experiment : str
            name of experiment (only relevant if this dataset contains more 
            than one experiment)
        vert_which : str or dict, optional
            valid AeroCom vertical info string encoded in name (e.g. Column,
            ModelLevel) or dictionary containing var_name as key and vertical
            coded string as value, accordingly
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case 
            the input ts_type is not available for this variable
        prefer_longer : bool
            if True and applicable, the ts_type resulting in the longer time
            coverage will be preferred over other possible frequencies that 
            match the query.
        require_all_vars_avail : bool
            if True, it is strictly required that all input variables are 
            available. 
        **kwargs
            optional and support for deprecated input args
        
        Returns
        -------
        tuple
            loaded data objects (type :class:`GriddedData`)
            
        Raises 
        ------
        IOError
            if input variable names is not list or string
        VarNotAvailableError
            1. if ``require_all_vars_avail=True`` and one or more of the 
            desired variables is not available in this class
            2. if ``require_all_vars_avail=True`` and if none of the input 
            variables is available in this object
        """
        if vars_to_retrieve is None and 'var_names' in kwargs:
            const.print_log.warning(DeprecationWarning('Input arg var_names '
                                                       'is deprecated (but '
                                                       'still works). Please '
                                                       'use vars_to_retrieve '
                                                       'instead'))
            vars_to_retrieve = kwargs['var_names']
        if vars_to_retrieve is None:
            vars_to_retrieve = self.vars
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        elif not isinstance(vars_to_retrieve, list):
            raise IOError('Invalid input for vars_to_retrieve {}. Need string '
                          'or list of strings specifying var_names to load. '
                          'You may also leave it empty (None) in which case all '
                          'available variables are loaded'
                          .format(vars_to_retrieve))
        if require_all_vars_avail:
            if not all([var in self.vars_provided for var in vars_to_retrieve]):
                raise VarNotAvailableError('One or more of the specified vars '
                                        '({}) is not available in {} database. '
                                        'Available vars: {}'.format(
                                        vars_to_retrieve, self.data_id, 
                                        self.vars_provided))
        var_names = list(np.intersect1d(self.vars_provided, vars_to_retrieve))
        if len(var_names) == 0:
            raise VarNotAvailableError('None of the desired variables is '
                                        'available in {}'.format(self.data_id))
        data = []
        for var in var_names:
            try:
                data.append(self.read_var(var_name=var, 
                                          start=start, stop=stop, 
                                          ts_type=ts_type,
                                          experiment=experiment,
                                          vert_which=vert_which,
                                          flex_ts_type=flex_ts_type,
                                          prefer_longer=prefer_longer))
            except (VarNotAvailableError, DataCoverageError) as e:
                self.logger.warning(repr(e))
        return tuple(data)
    
    def _check_correct_units_cube(self, cube):
        if ('invalid_units' in cube.attributes and 
            cube.attributes['invalid_units'] in const.GRID_IO.UNITS_ALIASES):
            
            from_unit = cube.attributes['invalid_units']
            to_unit = const.GRID_IO.UNITS_ALIASES[from_unit]
            const.logger.info('Updating invalid unit in {} from {} to {}'
                              .format(repr(cube), from_unit, to_unit))
            
            cube.units = to_unit
        return cube
                
    def _load_files(self, files, var_name, perform_fmt_checks=None,
                    **kwargs):
        """Load list of files containing variable to read into Cube instances
        
        Parameters
        ----------
        files : list
            list of netcdf file
        var_name : str
            name of variable to read
        perform_fmt_checks : bool
            if True, the loaded data is checked for consistency with 
            AeroCom default requirements.
        **kwargs
            additional keyword args parsed to :func:`load_cubes_custom`.
        
        Returns
        -------
        CubeList
            list of loaded Cube instances
        list
            list containing corresponding filenames of loaded cubes
        """
        cubes, loaded_files = load_cubes_custom(files, var_name,
                                                perform_fmt_checks=perform_fmt_checks,
                                                **kwargs)
        for cube in cubes:
            cube = self._check_correct_units_cube(cube)
            
        if len(loaded_files) == 0:
            raise IOError("None of the input files could be loaded in {}"
                          .format(self.data_id))
        
        self.loaded_cubes[var_name] = cubes
        return (cubes, loaded_files)
    
    def _load_var(self, var_name, ts_type, start, stop,
                  experiment, vert_which, flex_ts_type, 
                  prefer_longer, **kwargs):
        """Find files corresponding to input specs and load into GriddedData
        
        Note 
        ----
        See :func:`read_var` for I/O info.
        """
        if self.ignore_vert_code:
            vert_which = None
        subset = self.filter_query(var_name, ts_type, start, stop, 
                                   experiment, vert_which, 
                                   is_at_stations=False, 
                                   flex_ts_type=flex_ts_type, 
                                   prefer_longer=prefer_longer)
        if len(subset) == 0:
            raise DataQueryError('Could not find file match for query')
        ts_types = subset.ts_type.unique()
        # sanity check
        if len(ts_types) > 1:
            raise DataQueryError('Fatal: subset contains more than one ts_type')
        
        ts_type = ts_types[0]
        match_files = self._generate_file_paths(subset)
        (cube_list, 
         from_files) = self._load_files(match_files, var_name, **kwargs)
        is_concat = False
        if len(cube_list) > 1:
            try:
                cube = self.concatenate_cubes(cube_list)
                is_concat = True
            except iris.exceptions.ConcatenateError as e:
                raise NotImplementedError('Failed to concatenate cubes: {}\n'
                                          'Error: {}'.format(cube_list, repr(e)))
        else:
            cube = cube_list[0]
        
        data = GriddedData(input=cube, 
                           from_files=from_files,
                           data_id=self.data_id, 
                           ts_type=ts_type,
                           concatenated=is_concat)
        
        # crop cube in time (if applicable)
        if not start == 9999:
            try:
                data = self._check_crop_time(data, start, stop)
            except:
                const.print_log.exception('Failed to crop time dimension in {}. '
                                          '(start: {}, stop: {})'
                                          .format(data, start, stop))
        return data
    
    def _check_crop_time(self, data, start, stop):
        crop_time = False
        crop_time_range = [self.start, self.stop]
        if start is not None:
            crop_time = True
            crop_time_range[0] = to_pandas_timestamp(start)
        if stop is not None:
            crop_time = True
            crop_time_range[1] = to_pandas_timestamp(stop)
        if crop_time:
            self.logger.info("Applying temporal cropping of result cube")
            data = data.crop(time_range=crop_time_range)
        return data
    
    def _check_ts_type(self, ts_type):
        """Check and, if applicable, update ts_type
        
        Returns
        -------
        str
            valid ts_type
        
        Raises
        ------
        ValueError
            
        """
        if ts_type is None:
            if len(self.ts_types) == 0:
                raise AttributeError('Apparently no files with a valid ts_type '
                                     'entry in their filename could be found')
                
            ts_type = self.ts_types[0]
        if not ts_type in self.TS_TYPES:
            raise ValueError("Invalid input for ts_type, got: {}, "
                             "allowed values: {}".format(ts_type, 
                                                         self.TS_TYPES))
        return ts_type
    
    def __getitem__(self, var_name):
        """Try access import result for one of the models
        
        Parameters
        ----------
        var_name : str
            string specifying variable that is supposed to be extracted
        
        Returns
        -------
        GriddedData
            the corresponding read class for this model
            
        Raises
        -------
        ValueError
            if results for ``var_name`` are not available
        """
        if not var_name in self.data:
            self.read_var(var_name)
        return self.data[var_name]
    
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = ("\n{}\n{}\n"
             "Data ID: {}\n"
             "Data directory: {}\n"
             "Available experiments: {}\n"
             "Available years: {}\n"
             "Available frequencies {}\n"
             "Available variables: {}\n".format(head, 
                                                len(head)*"-",
                                                self.data_id,
                                                self.data_dir,
                                                self.experiments,
                                                self.years_avail,
                                                self.ts_types,
                                                self.vars))
        if self.data:
            s += "\nLoaded GriddedData objects:\n"
            for var_name, data in self.data.items():
                s += "{}\n".format(data.short_str())
# =============================================================================
#         if self.data_yearly:
#             s += "\nLoaded GriddedData objects (individual years):\n"
#             for var_name, yearly_data in self.data_yearly.items():
#                 if yearly_data:
#                     for year, data in yearly_data.items():
#                         s += "{}\n".format(data.short_str())
# =============================================================================
        return s.rstrip()
    
    ### DEPRECATED STUFF
    @property
    def name(self):
        """Deprecated name of attribute data_id"""
        const.print_log.warning(DeprecationWarning("Please use data_id"))
        return self.data_id
        
class ReadGriddedMulti(object):
    """Class for import of AEROCOM model data from multiple models
    
    This class provides an interface to import model results from an arbitrary
    number of models and specific for a certain time interval (that can be 
    defined, but must not be defined). Largely based on 
    :class:`ReadGridded`.
    
    ToDo
    ----
    
    Sub-class from ReadGridded
    
    Note
    ----
    The reading only works if files are stored using a valid file naming 
    convention. See package data file `file_conventions.ini <http://
    aerocom.met.no/pyaerocom/config_files.html#file-conventions>`__ for valid
    keys. You may define your own fileconvention in this file, if you wish.
    
    Attributes
    ----------
    data_ids : list
        list containing string IDs of all models that should be imported
    results : dict
        dictionary containing :class:`ReadGridded` instances for each
        name
    
    Examples
    --------
    >>> import pyaerocom, pandas
    >>> start, stop = pandas.Timestamp("2012-1-1"), pandas.Timestamp("2012-5-1")
    >>> models = ["AATSR_SU_v4.3", "CAM5.3-Oslo_CTRL2016"]
    >>> read = pyaerocom.io.ReadGriddedMulti(models, start, stop)
    >>> print(read.data_ids)
    ['AATSR_SU_v4.3', 'CAM5.3-Oslo_CTRL2016']
    >>> read_cam = read['CAM5.3-Oslo_CTRL2016']
    >>> assert type(read_cam) == pyaerocom.io.ReadGridded
    >>> for var in read_cam.vars: print(var)
    abs550aer
    deltaz3d
    humidity3d
    od440aer
    od550aer
    od550aer3d
    od550aerh2o
    od550dryaer
    od550dust
    od550lt1aer
    od870aer
    """

    def __init__(self, data_ids):
        const.print_log.warning(DeprecationWarning('ReadGriddedMulti class is '
                                                   'deprecated and will not '
                                                   'be further developed. '
                                                   'Please use ReadGridded.'))
        if isinstance(data_ids, str):
            data_ids = [data_ids]
        if not isinstance(data_ids, list) or not all([isinstance(x, str) for x in data_ids]):
            raise IllegalArgumentError("Please provide string or list of strings")
    
        self.data_ids = data_ids
        #: dictionary containing instances of :class:`ReadGridded` for each
        #: datset
        
        self.readers = {}
        self.data = {}
        
        self._init_readers()
        
    def _init_readers(self):
        for data_id in self.data_ids:
            self.readers[data_id] = ReadGridded(data_id)
        
    def read(self, vars_to_retrieve, start=None, stop=None,
             ts_type=None, **kwargs):
        """High level method to import data for multiple variables and models
        
        Parameters
        ----------
        var_names : :obj:`str` or :obj:`list`
            string IDs of all variables that are supposed to be imported
        start : :obj:`Timestamp` or :obj:`str`, optional
            start time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        stop : :obj:`Timestamp` or :obj:`str`, optional
            stop time of data import (if valid input, then the current 
            :attr:`start` will be overwritten)
        ts_type : str
            string specifying temporal resolution (choose from 
            "hourly", "3hourly", "daily", "monthly").If None, prioritised 
            of the available resolutions is used
        flex_ts_type : bool
            if True and if applicable, then another ts_type is used in case 
            the input ts_type is not available for this variable
        
        Returns
        -------
        dict
            result dictionary
            
        Examples
        --------
        
            >>> read = ReadGriddedMulti(names=["ECMWF_CAMS_REAN",
            ...                                "ECMWF_OSUITE"])
            >>> read.read(["od550aer", "od550so4", "od550bc"])
            
        """
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
            
        for data_id in self.data_ids:
            if not data_id in self.readers:
                self.readers[data_id] = ReadGridded(data_id)
            reader = self.readers[data_id]
            if not data_id in self.data:
                self.data[data_id] = {}
            for var in vars_to_retrieve:
                try:
                    data = reader.read_var(var, start, stop, ts_type, **kwargs)
                    self.data[data_id][var] = data
                except Exception as e:
                    const.print_log.exception('Failed to read data of {}\n'
                                            'Error message: {}'.format(data_id,
                                                                       repr(e)))
        return self.data
    
    def __str__(self):
        head = "Pyaerocom %s" %type(self).__name__
        s = ("\n%s\n%s\n"
             "Data-IDs: %s\n" %(head, len(head)*"-", self.data_ids))
        if bool(self.data):
            s += "\nLoaded data:"
            for name, vardata in self.data.items():
                for var, data in vardata.items():
                    s += "\n%s" %var
        return s
    
if __name__=="__main__":
    
    import matplotlib.pyplot as plt
    plt.close('all')
    import pyaerocom as pya
    
    r = ReadGridded('CAM6-Oslo_NHIST_f19_tn14_20190710_2010')
    
    data0 = r.read_var('od550gt1aer')
    
    fun = pya.io.aux_read_cubes.add_cubes
    
    data1 = r.read_var('od550gt1aer', aux_vars=['od550dust', 'od550ss'],
                       aux_fun=fun)
    
    
    data0.quickplot_map()
    data1.quickplot_map()
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  9 14:14:29 2018
"""
import numpy.testing as npt
import numpy as np
import os
from pyaerocom.test.settings import TEST_RTOL, lustre_unavail
from pyaerocom.io.read_aeronet_sunv2 import ReadAeronetSunV2
    
@lustre_unavail
def test_load_berlin_AeroSunV2L2D():
    reader = ReadAeronetSunV2()
    files = reader.find_in_file_list('*Berlin*')
    assert len(files) == 1
    assert os.path.basename(files[0]) == '920801_180519_Berlin_FUB.lev20'
    data = reader.read_file(files[0],
                            vars_to_retrieve=['od550aer'])
    
    test_vars = ['od440aer',
                 'od500aer',
                 'od550aer',
                 'ang4487aer']
    
    assert all([x in data for x in test_vars])
    
    # more than 100 timestamps
    assert all([len(data[x]) > 100 for x in test_vars])
    
    assert isinstance(data['dtime'][0], np.datetime64)
    assert data['dtime'][0] == np.datetime64('2014-07-06T00:00:00')
    
    first_vals = [data[var][0] for var in test_vars]
    
    nominal = [0.229427, 0.18302 , 0.151227, 2.002052]
    npt.assert_allclose(actual=first_vals, desired=nominal, rtol=TEST_RTOL)
    
    
    
    
if __name__=="__main__":
    test_load_berlin_AeroSunV2L2D()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Development template for reading routine of AirBase data.

NOTE
----

At the moment, the code below is just to illustrate the inheritance example 
from ReadUngriddedBase class. The code is not functional at the moment.

"""

from pyaerocom.io.readungriddedbase import ReadUngriddedBase

class ReadAirBase(ReadUngriddedBase):
    """Reading class for AirBase data"""
    DATASET_PATH = '/lustre/storeA/project/aerocom/aerocom1/AEROCOM_OBSDATA/AIRBASE/'
    DATA_ID = 'AirBase'
    _FILEMASK = '*.nc'
    
    DEFAULT_VARS = ['pm10']
    PROVIDES_VARIABLES = ['pm10']
    SUPPORTED_DATASETS = ['AirBase']
    TS_TYPE = None
    __version__ = '0.0.1'
    
    def read(self):
        pass
    
    def read_file(self):
        pass
    
if __name__=='__main__':
    
    r = ReadAirBase()
    
    
    files = r.get_file_list()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
High level test that checks the revision dates of all observation datasets
that are registered in ReadUngridded factory class

@author: jonasg
"""
import pytest
import numpy.testing as npt
from pyaerocom.io import ReadUngridded

from pyaerocom.test.settings import lustre_unavail

@lustre_unavail
def test_revision_dates():
    f = ReadUngridded()
    l = {}
    for s in f.SUPPORTED:
        r = s()
        l[r.DATA_ID] = r.data_revision
    res = []
    order = sorted(list(l.keys()))
    for name in order:
        res.append([name, l[name]])

    npt.assert_array_equal(res, 
                           [['AeronetInvV2Lev2.daily', '20171216'], 
                            ['AeronetInvV3Lev2.daily', '20190914'], 
                            ['AeronetSDAV2Lev2.daily', '20180519'], 
                            ['AeronetSDAV3Lev2.daily', '20190920'], 
                            ['AeronetSunV2Lev2.daily', '20180519'], 
                            ['AeronetSunV3Lev2.daily', '20190920'], 
                            ['DMS_AMS_CVO', '20190807'], 
                            ['EARLINET', '20190129'], 
                            ['EBASMC', '20191001'], 
                            ['GAWTADsubsetAasEtAl', '20190522']])

    
if __name__=="__main__":
    import sys
    pytest.main(sys.argv)
    #stat.plot_timeseries('scatc550aer')